"""Inbound message handler."""

from abc import ABCMeta

from ..address import Address
from ..constants import MessageFlagType
from ..subscriber_base import SubscriberBase
from ..utils import build_topic


class InboundHandlerBase(SubscriberBase):
    """Inbound message handler."""

    __meta__ = ABCMeta

    def __init__(self, topic, address=None, group=None, message_type=None):
        """Init the InboundHandlerBase class."""

        self._topic = topic
        if message_type is not None:
            self._message_type = MessageFlagType(message_type)
        else:
            self._message_type = None
        self._group = int(group) if group is not None else None
        self._address = Address(address) if address is not None else None
        subscriber_topic = build_topic(
            prefix="handler",
            topic=topic,
            address=address,
            group=group,
            message_type=message_type,
        )

        super().__init__(subscriber_topic=subscriber_topic)
        for attr_str in dir(self):
            attr = getattr(self, attr_str)
            if hasattr(attr, "register_handler"):
                attr.register_handler(
                    func=attr,
                    topic=topic,
                    address=address,
                    group=group,
                    message_type=message_type,
                )
            if hasattr(attr, "register_status"):
                # pylint: disable=no-member
                attr.register_status(attr, self._address.id)



"""Constants used by device groups and events."""

ON_INBOUND = "on_inbound"
OFF_INBOUND = "off_inbound"
ON_FAST_INBOUND = "on_fast_inbound"
OFF_FAST_INBOUND = "off_fast_inbound"
ON_ALL_LINK_CLEANUP = "on_all_link_cleanup"
OFF_ALL_LINK_CLEANUP = "off_all_link_cleanup"
OPEN_INBOUND = "open_inbound"
CLOSE_INBOUND = "close_inbound"

ON_HEARTBEAT_INBOUND = "on_heartbeat_inbound"
OFF_HEARTBEAT_INBOUND = "off_heartbeat_inbound"

STATUS_COMMAND = "status_command"

ON_COMMAND = "on_command"
OFF_COMMAND = "off_command"
ON_FAST_COMMAND = "on_fast_command"
OFF_FAST_COMMAND = "off_fast_command"

GET_OPERATING_FLAGS_COMMAND = "get_operating_flags"
SET_OPERATING_FLAGS_COMMAND = "set_operating_flags"
EXTENDED_GET_COMMAND = "extended_get"
EXTENDED_SET_COMMAND = "extended_set"
EXTENDED_GET_RESPONSE = "extended_get_response"

SET_LEDS_COMMAND = "set_leds"
TRIGGER_SCENE_ON_COMMAND = "trigger_scene_on"
TRIGGER_SCENE_OFF_COMMAND = "trigger_scene_off"

GET_IM_CONFIG_COMMAND = "get_im_config"
SET_IM_CONFIG_COMMAND = "set_im_config"

MANUAL_CHANGE = "manual_change"
ON_AT_RAMP_RATE = "on_at_ramp_rate"
OFF_AT_RAMP_RATE = "off_at_ramp_rate"



"""Command line tools to interact with the Insteon devices."""
import asyncio
from binascii import Error as BinasciiError, unhexlify
from functools import partial
import os

from .. import async_close, async_connect, devices
from ..address import Address
from ..constants import HC_LOOKUP, UC_LOOKUP
from ..managers.link_manager import async_cancel_linking_mode, async_unlink_devices
from .aldb import ToolsAldb
from .commands import ToolsCommands
from .config import ToolsConfig
from .scenes import ToolsScenes
from .tools_base import DEFAULT_HUB_PORT, ToolsBase

DEFAULT_HUB_VERSION = 2


async def _add_device_with_yield(log_stdout, address: Address, multiple: bool):
    """Act as a frontend to devices.async_add_device to capture the yielding of addresses."""
    async for address in devices.async_add_device(address=address, multiple=multiple):
        log_stdout(f"Device {str(address)} was added.")


class InsteonCmd(ToolsBase):
    """Command class to test interactivity."""

    def __init__(self, loop, args=None, menu=None, stdin=None, stdout=None):
        """Init the InsteonCmd class."""
        super().__init__(loop, args=args, menu=menu, stdin=stdin, stdout=stdout)
        self._link_mode_lock = asyncio.Lock()
        self._link_queue = asyncio.Queue()

    async def do_connect(
        self,
        device_or_host,
        username=None,
        hub_version=None,
        port=None,
        password=None,
    ):
        """Connect to the Insteon modem.

        Usage:
            connect
            connect device
            connect host username [hub_version [port [password]]]
        """
        if not device_or_host:
            self.device = None
            self.host = None
            self.username = None
            self.hub_version = None
            self.port = None
            password = None
            password = await self._get_connection_params(password)

        elif username is not None:
            self.host = device_or_host
            self.username = username
            if hub_version is not None:
                try:
                    self.hub_version = int(hub_version)
                except ValueError:
                    self._log_stdout("Invalid Hub version")
                    self.hub_version = None
            else:
                self.hub_version = DEFAULT_HUB_VERSION
            if port is not None:
                self.port = port
            elif self.hub_version in [1, 2]:
                self.port = DEFAULT_HUB_PORT[self.hub_version]
            else:
                self.port = None
            if password is None:
                password = await self._get_connection_params(password)
        elif device_or_host == "mock":
            self.host = "127.0.0.1"
            self.port = 8080
            self.mock = True
        else:
            self.device = device_or_host

        if not self.device and not self.host:
            self._log_stdout("Modem information required.")
            return

        try:
            await async_connect(
                device=self.device,
                host=self.host,
                port=self.port,
                username=self.username,
                password=password,
                hub_version=self.hub_version,
                mock=self.mock,
            )
            connected_to = self.device if self.device else self.host
            self._log_stdout(f"Connected to {connected_to}")
        except ConnectionError:
            self._log_stdout("Connection failed. Please review connection information.")

    async def do_disconnect(self):
        """Close the connection to the modem.

        Usage:
            disconnect
        """
        await async_close()

    async def do_load_devices(
        self,
        workdir=None,
        id_devices=1,
        load_modem_aldb=1,
        log_stdout=None,
        background=False,
    ):
        """Load the devices.

        Usage:
            load_devices [--background | -b] workdir id_devices load_modem_aldb

        workdir: Directory where the saved device file is located (Default is current directory)
        id_devices: Option for handling unknown devices
            0 - Do not ID devices
            1 - ID unknown devices only (default)
            2 - ID all devices
        load_modem_aldb: Option for loading the Modem ALDB
            0 - Do not load
            1 - Load if not loaded from saved file (default)
            2 - Always load
        """

        self.workdir = workdir
        if self.workdir == ".":
            self.workdir = os.getcwd()

        if self.workdir:
            if not os.path.exists(self.workdir):
                log_stdout("Invalid working directory")
                self.workdir = None

        if not self.workdir:
            if background:
                log_stdout("Working directory is required in background mode.")
                return
            self.workdir = await self._get_workdir()

        try:
            id_devices = await self._ensure_int(
                id_devices,
                [0, 1, 2],
                "ID Devices",
                ask_value=not background,
                log_stdout=log_stdout,
                default=1,
            )
        except ValueError:
            log_stdout("Invalid value for `id_devices`")
            return

        try:
            load_modem_aldb = await self._ensure_int(
                load_modem_aldb,
                [0, 1, 2],
                "Load modem ALDB",
                ask_value=not background,
                log_stdout=log_stdout,
                default=1,
            )
        except ValueError:
            log_stdout("Invalid value for `load_modem_aldb`")
            return

        await devices.async_load(
            workdir=self.workdir, id_devices=id_devices, load_modem_aldb=load_modem_aldb
        )
        log_stdout(f"Total devices: {len(devices)}")

    async def menu_aldb(self):
        """Manage device All-Link database."""
        await self._call_next_menu(ToolsAldb, "aldb")

    async def menu_config(self):
        """Manage operational flags and extended properties."""
        await self._call_next_menu(ToolsConfig, "config")

    async def menu_commands(self):
        """Execute device commands."""
        await self._call_next_menu(ToolsCommands, "commands")

    async def menu_scenes(self):
        """Manage scenes."""
        await self._call_next_menu(ToolsScenes, "scenes")

    async def do_add_device(self, address_or_multiple=None):
        """Link a device to the modem.

        Usage:
            add_device [<address> | multiple | m]

        <address> (Optional): Insteon address of the device to link
        multiple | m (optional): Add multiple devices at in one session.

        If an address is not provided the device must be put into link mode manually.
        This is done by pressing the set button on the device.

        If the word `multiple` or `m` are provided then the modem will stay in linking mode until
        the `cancel_lihnking` command is entered at the command line or for 3 minutes after
        the last device is linked.

        NOTE: Not all devices respond to a request to link using thier address. For
        any device that does not respond, the device must be put into linking mode
        manually.

        This command cannot be run in the background.
        """
        multiple = address_or_multiple is not None and address_or_multiple[:1] == "m"
        address = None

        if address_or_multiple is None:
            multiple = await self._ensure_bool(
                None,
                "Multiple devices",
                ask_value=True,
                log_stdout=self._log_stdout,
                default="n",
            )

        if not multiple:
            addresses = await self._ensure_address(
                address_or_multiple,
                "Address (optional)",
                ask_value=True,
                log_stdout=self._log_stdout,
                allow_all=False,
                match_device=False,
            )
            if addresses:
                address = addresses[0]

        if address is not None:
            self._log_stdout(f"Attempting to place device {address} into linking mode.")
            self._log_stdout(
                "If the device does not link with the modem within a minute, press the set button on the device."
            )
        else:
            self._log_stdout(
                "Press the set button on the device. Linking will occur in the background."
            )

        self._log_stdout("Press enter to stop linking.")
        link_function = partial(
            _add_device_with_yield,
            log_stdout=self._log_stdout,
            address=address,
            multiple=multiple,
        )
        try:
            await self._start_all_linking(link_function)
        except asyncio.TimeoutError:
            self._log_stdout("All-Linking has timed out.")

    async def do_add_device_manually(
        self, address, cat, subcat, firmware="0x00", log_stdout=None, background=None
    ):
        """Add a device using a cat and subcat.

        Usage:
           add_device_manually [--background | -b] address cat subcat [firmware]

        address: Device address
        cat: Device category (i.e. 05 or 0x05)
        subcat: Device subcategory (i.e. 1f or 0x1f)

        Note: Device cat and subcat are the hex representation and should be at
        least 2 digits. They can be written with or without the hex prefix 0x.

        The cat and subcat can be found in the user manual for the device. The values
        in the user manual are in hex.
        """

        def hex_str_to_int(value):
            """Convert a hex string into an integer."""
            if value[0:2] == "0x":
                value = value[2:]
            byte_val = unhexlify(value)
            return int.from_bytes(byte_val, byteorder="big")

        if subcat is None and not background:
            firmware = None

        try:
            addresses = await self._ensure_address(
                address,
                "Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=False,
            )
            if not addresses:
                raise ValueError
            address = addresses[0]
        except (IndexError, ValueError):
            log_stdout("Address required.")
            return

        while True:
            if cat is not None:
                try:
                    cat = hex_str_to_int(cat)
                    break
                except (BinasciiError, TypeError, IndexError, ValueError):
                    if background:
                        log_stdout("Invalid device category.")
                        return
                    cat = None
            cat = await self._input("Device category or enter to cancel (eg: 0x02)")
            if cat == "":
                return

        while True:
            if subcat is not None:
                try:
                    subcat = hex_str_to_int(subcat)
                    break
                except (BinasciiError, TypeError, IndexError, ValueError):
                    if background:
                        log_stdout("Invalid device subcategory.")
                        return
                    subcat = None
            subcat = await self._input(
                "Device subcategory or enter to cancel (eg: 0x02)"
            )
            if subcat == "":
                return

        while True:
            if firmware is not None:
                try:
                    firmware = hex_str_to_int(firmware)
                except (BinasciiError, TypeError, IndexError, ValueError):
                    if background:
                        log_stdout("Invalid device firmware.")
                        return
                    firmware = None
                else:
                    break
            firmware = await self._input(
                "Device firmware or enter to cancel (eg: 0x02)"
            )
            if firmware == "":
                return

        self._log_background(f"Calling with {address} {cat} {subcat} {firmware}")

        devices.set_id(address, cat, subcat, firmware)

    async def do_remove_device(
        self, address=None, force=None, log_stdout=None, background=False
    ):
        """Remove a device.

        Usage:
          remove_device [--background | -b] [address [f|force]]

        address (optional): Device address
        f or force (optional): Force removal.

        This command will perform the following steps:
          - Remove all modem links from the device
          - Unlink the device from the modem
          - Remove all device links from the modem

        The force option is used when the device is no longer available, for example, if it has been removed from the network.
        If the force option is used,  all links will be removed from the modem. This will not attempt to unlink the device or
        remove modem links from the device.
        """

        try:
            addresses = await self._ensure_address(
                address,
                "Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=False,
            )
            if not addresses:
                log_stdout("Device address is required.")
                return
            address = addresses[0]
        except (IndexError, ValueError):
            log_stdout("Invalid Insteon address.")
            return

        device = devices[address]
        if device is None and force is None:
            log_stdout(
                f"Device with address {address} not found. Use force mode to remove links from the modem."
            )
            if background:
                return

        if device is not None and force is None:
            force = ""

        force = await self._ensure_bool(
            force[:1] if force is not None else None,
            "Force",
            ask_value=not background,
            log_stdout=log_stdout,
            true_val="f",
            values=["f", ""],
            default=False,
        )

        if force:
            await async_unlink_devices(devices.modem, address)
            devices[address] = None
            return

        if device is None:
            log_stdout(
                f"Device with address {address} not found. Use force mode to remove links from the modem."
            )
            return

        link_function = partial(devices.async_remove_device, address=address)
        try:
            linked_address = await self._start_all_linking(
                link_function=link_function, background=background
            )
            if linked_address:
                log_stdout(f"Device {linked_address} was removed.")
        except asyncio.TimeoutError:
            log_stdout("No device was removed due to a timeout error.")

    async def do_add_x10_device(
        self, housecode, unitcode, x10_type, steps=22, log_stdout=None, background=False
    ):
        """Add an X10 device.

        Usage:
            add_x10_device [--background | -b] housecode unitcode type [steps]

        houscode: The device housecode (a - p)
        unitcode: The device unitcode (1 - 16)
        type: The device type (on_off, dimmable, sensor)
        steps: Number of steps from off to full on (dimmable only)
        """
        x10_types = ["dimmable", "on_off", "sensor"]
        if housecode is not None:
            housecode = housecode.lower()
        if x10_type is not None:
            x10_type = x10_type.lower()

        if background:
            log_stdout("In background mode")

        try:
            housecode = await self._ensure_string(
                housecode,
                list(HC_LOOKUP.keys()),
                "Housecode",
                ask_value=not background,
                log_stdout=log_stdout,
            )
            if housecode is None:
                return
        except ValueError:
            if background:
                log_stdout("Housecode is required")
            return

        try:
            unitcode = await self._ensure_int(
                unitcode,
                list(UC_LOOKUP.keys()),
                "Unitcode",
                ask_value=not background,
                log_stdout=log_stdout,
            )
            if unitcode is None:
                return
        except ValueError:
            if background:
                log_stdout("Unitcode is required")
            return

        ask_steps = x10_type is None

        try:
            x10_type = await self._ensure_string(
                x10_type,
                x10_types,
                "X10 device type",
                ask_value=not background,
                log_stdout=log_stdout,
            )
            if x10_type is None:
                return
        except ValueError:
            if background:
                log_stdout("X10 devuce type is required")
            return

        if x10_type == "dimmable" and ask_steps:
            steps = await self._ensure_byte(
                None,
                "Dimmer steps",
                ask_value=not background,
                log_stdout=log_stdout,
                default=22,
            )

        devices.add_x10_device(housecode, unitcode, x10_type, steps, 255)

    async def _start_all_linking(self, link_function, background=False):
        """Put the modem into linking mode repeatedly or cancel on timeout."""
        unfinished = []
        result = ""
        if background:
            return await link_function()

        try:
            link_func_task = asyncio.create_task(link_function())
            input_task = asyncio.create_task(self._input())
            finished, unfinished = await asyncio.wait(
                [link_func_task, input_task],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for task in finished:
                return task.result()
        finally:
            for task in unfinished:
                task.cancel()
            if unfinished:
                await asyncio.wait(unfinished)
            if result == "":
                self._log_command("Sending cancel command")
                await async_cancel_linking_mode()
            await asyncio.sleep(0.1)


def tools():
    """Start insteon tools."""
    InsteonCmd.start()



"""Command handler for message ID 0x6f Manage All-Link Record."""

from . import ack_handler, nak_handler
from ..address import Address
from ..constants import ManageAllLinkRecordAction, ResponseStatus
from ..data_types.all_link_record_flags import AllLinkRecordFlags
from ..topics import MANAGE_ALL_LINK_RECORD
from .outbound_base import OutboundHandlerBase


class ManageAllLinkRecordCommand(OutboundHandlerBase):
    """Command handler for message ID 0x6F Manage All-Link Record."""

    def __init__(self):
        """Init the ManageAllLinkRecordCommand class."""
        super().__init__(topic=MANAGE_ALL_LINK_RECORD)

    # pylint: disable=arguments-differ
    async def async_send(
        self,
        action: ManageAllLinkRecordAction,
        controller: bool,
        group: int,
        target: Address,
        data1: int = 0,
        data2: int = 0,
        data3: int = 0,
        in_use: bool = True,
        high_water_mark: bool = False,
        bit5: bool = True,
        bit4: bool = False,
    ):
        """Send Manage All-Link Record command."""
        flags = AllLinkRecordFlags.create(
            in_use=in_use,
            controller=controller,
            hwm=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
        return await super().async_send(
            action=action,
            flags=flags,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
        )

    @ack_handler
    async def async_handle_ack(self, action, flags, group, target, data1, data2, data3):
        """Handle ACK response."""
        await self._async_handle_ack()

    @nak_handler
    async def async_handle_nak(self, action, flags, group, target, data1, data2, data3):
        """Handle NAK response."""
        await self._message_response.put(ResponseStatus.FAILURE)



"""Tools to run device commands."""

import inspect
from types import MethodType

from .. import (
    async_x10_all_lights_off,
    async_x10_all_lights_on,
    async_x10_all_units_off,
    devices,
)
from ..constants import HC_LOOKUP
from ..managers.scene_manager import async_trigger_scene_off, async_trigger_scene_on
from .tools_base import ToolsBase


class ToolsCommands(ToolsBase):
    """Tools to run device commands."""

    async def do_on(self, address, log_stdout=None, background=False):
        """Test turning a device on.

        Usage:
            on [--background | -b] <address> [group]

        This method calls `async_on` with default values.
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]
        if not hasattr(device, "async_on"):
            log_stdout("Device cannot be turned on.")
            return

        await device.async_on()

    async def do_off(self, address, log_stdout=None, background=False):
        """Test turning a device off.

        Usage:
            on [--background | -b] <address> [group]

        This method calls `async_off` with default values.
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]
        if not hasattr(device, "async_off"):
            log_stdout("Device cannot be turned off.")
            return

        await device.async_off()

    async def do_scene_on(self, scene, log_stdout=None, background=False):
        """Test a scene.

        Usage:
            scene_on [--background | -b] scene
        """

        try:
            scene = await self._ensure_byte(scene, "Scene", not background, log_stdout)
            if scene is None:
                log_stdout("Scene number is required")
                return
        except ValueError:
            log_stdout("Invalid scene number")
            return

        await async_trigger_scene_on(scene)

    async def do_scene_off(self, scene, log_stdout=None, background=False):
        """Test a scene.

        Usage:
            scene_off [--background | -b] scene
        """

        try:
            scene = await self._ensure_byte(scene, "Scene", not background, log_stdout)
            if scene is None:
                log_stdout("Scene number is required")
                return
        except ValueError:
            log_stdout("Invalid scene number")
            return

        await async_trigger_scene_off(scene)

    async def do_cmd(self, address, command):
        """Run a general device command.

        Usage:
            cmd address command

        Examples:
            cmd 1a2b3c async_on
            cmd 1a2b3c async_off

        You will be prompted for any arguments to the command. To see a list of available commands for a device use:
            help cmd address

        This command cannot be run in the background.
        """

        addresses = await self._ensure_address(
            address=address,
            name="Address",
            ask_value=True,
            log_stdout=self._log_stdout,
            allow_all=False,
            match_device=True,
        )
        if not addresses:
            self._log_stdout("Address is required")
            return

        address = addresses[0]
        device = devices[address]

        device_commands = []
        for device_method in dir(device):
            if not device_method.startswith("_"):
                device_commands.append(device_method)

        command = await self._ensure_string(
            command, device_commands, "Command", True, self._log_stdout
        )
        if not command:
            self._log_stdout("A device command is required")
            return

        func = getattr(device, command)
        signature = inspect.signature(func)
        func_args = [arg for arg, _ in signature.parameters.items()]
        kwargs = {}
        if func_args:
            args_defaults = {
                k: v.default
                for k, v in signature.parameters.items()
                if v.default is not inspect.Parameter.empty
            }
            print_args = ""
            for arg, _ in signature.parameters.items():
                if print_args != "":
                    print_args += ", "
                print_args += arg
                arg_def = args_defaults.get(arg, inspect.Parameter.empty)
                if arg_def != inspect.Parameter.empty:
                    print_args += f"={arg_def}"

            # self._log_stdout(f"Command arguments: {print_args}")
            self._log_stdout(f"Command arguments: {signature}")
            while True:
                arg = await self._ensure_string(
                    None, func_args, "Argument", True, self._log_stdout
                )
                if arg is None:
                    break
                arg_default = args_defaults.get(arg)
                value = await self._ensure_arg_value(
                    arg, None, True, self._log_stdout, arg_default
                )
                kwargs[arg] = value
            for arg in func_args:
                if arg not in kwargs and arg not in args_defaults:
                    self._log_stdout(f"Missing value for required argument {arg}")
                    return
                kwargs[arg] = kwargs.get(arg, args_defaults.get(arg))

        if inspect.iscoroutinefunction(func) or inspect.isawaitable(func):
            result = await func(**kwargs)
        else:
            result = func(**kwargs)
        self._log_stdout(f"Result: {str(result)}")

    async def do_x10_all_lights_on(
        self, housecode: str, log_stdout=None, background=False
    ):
        """Run the X10 All Lights On command.

        Usage:
          x10_all_lights_on housecode
        """
        housecode = await self._ensure_string(
            str(housecode).lower(),
            list(HC_LOOKUP.keys()),
            "Housecode",
            True,
            log_stdout,
        )
        await async_x10_all_lights_on(housecode=housecode)

    async def do_x10_all_lights_off(
        self, housecode: str, log_stdout=None, background=False
    ):
        """Run the X10 All Lights Off command.

        Usage:
          x10_all_lights_off housecode
        """
        housecode = await self._ensure_string(
            str(housecode).lower(),
            list(HC_LOOKUP.keys()),
            "Housecode",
            True,
            log_stdout,
        )
        await async_x10_all_lights_off(housecode=housecode)

    async def do_x10_all_units_off(
        self, housecode: str, log_stdout=None, background=False
    ):
        """Run the X10 All Units Off command.

        Usage:
          x10_all_units_off housecode
        """
        housecode = await self._ensure_string(
            str(housecode).lower(),
            list(HC_LOOKUP.keys()),
            "Housecode",
            True,
            log_stdout,
        )
        await async_x10_all_units_off(housecode=housecode)

    # pylint: disable=arguments-renamed
    # pylint: disable=invalid-overridden-method
    async def do_help(self, cmd=None, address=None, method=None):
        """List available commands with "help" or detailed help with "help cmd".

        You can also get a list of device commands to run with the "cmd" method.
        Example 1:
            help cmd 1a2b3c

        This will show the available commands that can be run against device 1A.2B.3C.

        Example 2:
            help cmd 1a2b3c async_off

        This will show the parameters available to the async_off method.
        """

        if cmd != "cmd":
            return super().do_help(cmd)

        try:
            addresses = await self._ensure_address(
                address,
                "Device address",
                False,
                self._log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                return super().do_help(cmd)

            address = addresses[0]
        except ValueError:
            self._log_stdout("Invalid device address")
            return

        device = devices[address]
        if method:
            if not hasattr(device, method):
                self._log_stdout(f"Method {method} not found for device {address}.")
                return
            func = getattr(device, method)
            doc = func.__doc__
            if doc:
                self._log_stdout(doc)
            arguments = str(inspect.signature(func)).strip("(").strip(")")
            self._log_stdout("")
            self._log_stdout(
                f"Device {str(device.address)} {method} method arguments: {arguments}"
            )
            return

        all_methods = dir(device)
        methods = []
        for func_method in all_methods:
            if func_method[0] != "_":
                func = getattr(device, func_method)
                if isinstance(func, MethodType):
                    methods.append(func_method)
        self._log_stdout("")
        header = f"Available commands for device {str(device.address)}:"
        self.print_topics(header, methods, 15, 80)



"""Modem command to get next ALDB record."""
import logging

from . import ack_handler, nak_handler
from ..address import Address
from ..constants import ResponseStatus
from ..data_types.all_link_record_flags import AllLinkRecordFlags
from ..topics import WRITE_EEPROM
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class WriteEepromHandler(OutboundHandlerBase):
    """Handle Write EEPROM commands."""

    def __init__(self):
        """Init the ReadEepromHandler class."""
        super().__init__(topic=WRITE_EEPROM)
        _LOGGER.debug("Setup WriteEepromHandler")

    # pylint: disable=arguments-differ
    async def async_send(
        self,
        mem_addr: int,
        in_use: bool,
        high_water_mark: bool,
        controller: bool,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
        bit5: bool,
        bit4: bool,
    ):
        """Send the Write to EEPROM message."""
        mem_array = mem_addr.to_bytes(2, "big")
        flags = AllLinkRecordFlags.create(
            in_use=in_use,
            controller=controller,
            hwm=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
        return await super().async_send(
            mem_hi=mem_array[0],
            mem_low=mem_array[1] - 7,
            flags=flags,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
        )

    @ack_handler
    async def async_handle_ack(
        self,
        mem_addr: int,
        flags: AllLinkRecordFlags,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
    ):
        """Send the Read from EEPROM message."""
        await self._async_handle_ack()

    @nak_handler
    async def async_handle_nak(
        self,
        mem_addr: int,
        flags: int,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
    ):
        """Receive the NAK message and return False."""
        await self._message_response.put(ResponseStatus.FAILURE)



"""On Leve state."""

import logging

from ..address import Address
from .group_base import GroupBase

_LOGGER = logging.getLogger(__name__)


class OnLevel(GroupBase):
    """Variable On Level state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: int = None,
        status_type: int = 0,
    ):
        """Init the OnLevel class."""
        super().__init__(
            name, address, group, default, value_type=int, status_type=status_type
        )
        self._is_dimmable: bool = True

    @property
    def value(self):
        """Return the value of the state."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the state."""
        try:
            value = self._type(value) if value is not None else None
            if not self._is_dimmable and value:
                value = 255
            if self._value == value:
                return
            self._value = value

        except TypeError as ex:
            raise TypeError(
                f"Error setting value of State {self._name}: Must be of type {self._type.__name__}"
            ) from ex
        self._call_subscribers(
            name=self._name,
            address=self._address.id,
            value=self._value,
            group=self._group,
        )

    # pylint: disable=arguments-differ
    def set_value(self, on_level: int):
        """Set the value of the state from the handlers."""
        self.value = on_level



"""Manage status requests.

Status requests should not run more often than necessary due to the nature of linking the status message response to the original request.
If status messages overlap, the response can be very difficult to associate to the original request. This is especially important for
devices that have multiple status messages.

Logic:
first call always happens when requested
second call happens
    if
        No other call is in progress then run
    else
        if
            Another call is in queue then cancel
        else
            Wait until the first call is done then run

"""

from asyncio import Lock
from logging import DEBUG, getLogger
from typing import Callable, Dict, Union

from ..address import Address
from ..constants import ResponseStatus
from ..handlers.to_device.status_request import StatusRequestCommand
from ..utils import multiple_status

_LOGGER = getLogger(__name__)
_LOGGER.setLevel(DEBUG)


class StatusManager:
    """Status manager."""

    def __init__(self, address: Address):
        """Init the StatusManager class."""
        self._address = Address(address)
        self._status_cmds: Dict[int, StatusRequestCommand] = {}
        self._callbacks: Dict[int, Callable] = {}
        self._call_waiting: bool = False
        self._run_lock = Lock()

    def add_status_type(self, status_type: int, callback_function: Callable):
        """Add a status request type."""
        self.remove_status_type(status_type=status_type)

        status_cmd = StatusRequestCommand(self._address, status_type)
        status_cmd.subscribe(callback_function)
        self._status_cmds[status_type] = status_cmd
        self._callbacks[status_type] = callback_function

    def remove_status_type(self, status_type: int):
        """Remove a status request type and unsubscribe to command responses."""
        status_cmd: StatusRequestCommand = self._status_cmds.get(status_type)
        if status_cmd:
            status_cmd.unsubscribe(self._callbacks[status_type])
            self._callbacks.pop(status_type)
            self._status_cmds.pop(status_type)

    async def async_status(
        self, status_type: Union[int, None] = None
    ) -> ResponseStatus:
        """Send the status request."""
        if self._call_waiting:
            # No need for this call because an existing call is already scheduled
            _LOGGER.debug("No need to run this status request.")
            return ResponseStatus.SUCCESS

        if self._run_lock.locked():
            # This is the second call and needs to wait for the first call to finish
            # Make sure any subsequent calls know there is already a call waiting
            _LOGGER.debug(
                "This status request is waiting for the prior request to finish."
            )
            self._call_waiting = True

        results = []
        if status_type is not None and status_type not in self._status_cmds:
            status_type = None
        status_types = (
            [status_type] if status_type is not None else self._status_cmds.keys()
        )
        async with self._run_lock:
            for curr_status_type in status_types:
                result = await self._async_status(self._status_cmds[curr_status_type])
                results.append(result)
            self._call_waiting = False

        return multiple_status(*results)

    async def _async_status(self, status_cmd: Callable) -> ResponseStatus:
        """Execute the status command."""
        retries = 2
        result = ResponseStatus.UNSENT
        while retries and result not in [
            ResponseStatus.SUCCESS,
            ResponseStatus.DIRECT_NAK_ALDB,
        ]:
            result = await status_cmd.async_send()
            retries -= 1
        return result



"""All-Link database for an Insteon Modem."""
import logging
from typing import List

from ..address import Address
from ..constants import ALDBStatus, EngineVersion, ReadWriteMode
from ..managers.aldb_im_write_manager import ImWriteManager
from .aldb_base import ALDBBase
from .aldb_record import ALDBRecord

_LOGGER = logging.getLogger(__name__)
MAX_RETRIES = 3


class MockModemALDB(ALDBBase):
    """Mock All-Link database for modems.

    Subscribed topics:
    modem.aldb.loaded: Triggered when the ALDB load command completes.

    Messages sent:
    modem.aldb.load: Triggers the loading of the ALDB.
    """

    def __init__(self, address, version=EngineVersion.I2, mem_addr=0x1FFF):
        """Init the ModemALDB."""

        super().__init__(address, version, mem_addr, write_manager=ImWriteManager)

        self._read_write_mode = ReadWriteMode.UNKNOWN
        self._read_manager = None
        self._default_records = {}

    @property
    def read_write_mode(self) -> ReadWriteMode:
        """Emit the modem read mode."""
        return self._read_write_mode

    @read_write_mode.setter
    def read_write_mode(self, value: ReadWriteMode):
        """Set the modem read mode."""
        self._read_write_mode = ReadWriteMode(value)

    # pylint: disable=arguments-differ
    async def async_load(self, *args, **kwargs):
        """Load the All-Link Database."""

        if self._read_manager is None:
            return

        self._update_status(ALDBStatus.LOADING)
        # See if we can use EEPROM reads
        if self._read_write_mode == ReadWriteMode.UNKNOWN:
            top_of_mem = await self._read_manager.async_confirm_eeprom_read()
            if top_of_mem:
                self._mem_addr = top_of_mem
                self._read_write_mode = ReadWriteMode.EEPROM
            else:
                self._read_write_mode = ReadWriteMode.STANDARD

        _LOGGER.debug("Loading the modem ALDB: %s", self._read_write_mode)
        if self._read_write_mode == ReadWriteMode.EEPROM:
            await self._async_load_eeprom()
        else:
            await self._async_load_standard()
        self.set_load_status()

        return self._status

    async def async_find_records(self, address: Address, group: int):
        """Find an All-Link record in the Modem ALDB."""
        async for rec in self._read_manager.async_find(address, group):
            yield rec

    async def async_read_record(self, mem_addr):
        """Read from EEPROM."""
        return await self._read_manager.async_read_record(mem_addr=mem_addr)

    def load_saved_records(
        self, status: ALDBStatus, records: List[ALDBRecord], first_mem_addr: int = None
    ):
        """Load All-Link records from a dictionary of saved records."""
        super().load_saved_records(status, records, first_mem_addr)
        self._default_records = self._records.copy()

    async def _async_load_standard(self):
        """Load using get first and get next methods."""
        next_mem_addr = self.first_mem_addr
        self._records = {}
        async for rec in self._read_manager.async_load_standard():
            rec.mem_addr = next_mem_addr
            self._records[next_mem_addr] = rec
            self._notify_change(rec)
            next_mem_addr -= 8

    async def _async_load_eeprom(self):
        """Load using EEPROM read method."""
        _LOGGER.debug("Loading from EEPROM")
        next_mem_addr = self.first_mem_addr
        self._records = {}
        record = await self._read_manager.async_read_record(next_mem_addr)
        while record:
            self._records[record.mem_addr] = record
            self._notify_change(record)
            if record.is_high_water_mark:
                return
            record = await self._read_manager.async_read_record(next_mem_addr)
            next_mem_addr -= 8

    def _is_loaded(self):
        """Calculate the AlDB load status."""
        if self._read_write_mode == ReadWriteMode.STANDARD:
            return ALDBStatus.LOADED
        return super()._is_loaded()



"""Command line interface for Operation Flag and Extended Properites Management."""

from .. import devices
from ..config.device_flag import DeviceFlagBase
from ..constants import PropertyType, RelayMode, ResponseStatus
from .tools_base import ToolsBase


def _show_flag(flag: DeviceFlagBase, advanced: bool, hidden: bool):
    if flag.property_type == PropertyType.STANDARD:
        return True
    if flag.property_type == PropertyType.ADVANCED and advanced:
        return True
    if flag.property_type == PropertyType.HIDDEN and hidden:
        return True
    return False


class ToolsConfig(ToolsBase):
    """Command line interface for Operation Flag and Extended Properites Management."""

    async def do_read_config(self, address, log_stdout=None, background=False):
        """Read the operating flags and extended properties of a device.

        Usage:
            read_config <ADDRESS>|all
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=True,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        for device_address in addresses:
            device = devices[device_address]
            if not device == devices.modem:
                log_stdout(f"Reading configuration for: {device.address}")
                if await device.async_read_op_flags() == ResponseStatus.SUCCESS:
                    log_stdout("Operating flags read")
                else:
                    log_stdout("Operating flags read error")
                if await device.async_read_ext_properties() == ResponseStatus.SUCCESS:
                    log_stdout("Extended properties read")
                else:
                    log_stdout("Extended properties read error")

    async def do_print_config(
        self, address, advanced="n", hidden="n", log_stdout=None, background=False
    ):
        """Write the operating flags and extended properties to a device.

        Usage:
            read_ops_flags <ADDRESS>|all  <advanced> <hidden>

            advanced: y|n default 'n'
            hidden: y|n default 'n'
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        advanced = await self._ensure_bool(
            advanced, "Advanced", ask_value=True, log_stdout=log_stdout
        )
        hidden = await self._ensure_bool(
            hidden, "Advanced", ask_value=True, log_stdout=log_stdout
        )

        for device_address in addresses:
            device = devices[device_address]
            if device and not device == devices.modem:
                log_stdout("")
                log_stdout(f"Device: {device_address}")
                log_stdout("Property                        Value")
                log_stdout("------------------------------  ----------")
                for name in device.operating_flags:
                    op_flag = device.operating_flags[name]
                    if _show_flag(op_flag, advanced, hidden):
                        if op_flag.is_dirty:
                            name_out = f"{name}*"
                        else:
                            name_out = name
                        if op_flag.value_type == bool:
                            log_stdout(f"{name_out:30s}  {bool(op_flag.value)}")
                        else:
                            prop_value = (
                                op_flag.value if op_flag.value is not None else 0
                            )
                            log_stdout(
                                f"{name_out:30s}  0x{prop_value:02x} ({prop_value})"
                            )
                log_stdout("")
                for name in device.properties:
                    prop = device.properties[name]
                    if _show_flag(prop, advanced, hidden):
                        if prop.is_dirty:
                            name_out = f"{name}*"
                        else:
                            name_out = name
                        if isinstance(prop.value, bool):
                            log_stdout(f"{name_out:30s}  {prop.value}")
                        else:
                            prop_value = prop.value if prop.value is not None else 0
                            log_stdout(
                                f"{name_out:30s}  0x{prop_value:02x} ({prop_value:d})"
                            )
                log_stdout("")

    async def do_set_config_value(
        self, address, name, value, log_stdout=None, background=False
    ):
        """Set an operating flag or extended property for a device.

        Usage:
            set_config_value <ADDRESS> flag y|n
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required.")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]

        prop_names = []
        for prop_name in device.operating_flags:
            prop_names.append(prop_name)
        for prop_name in device.properties:
            prop_names.append(prop_name)

        if not prop_names:
            log_stdout(f"Device {address} has no configurable settings.")
            return

        try:
            name = await self._ensure_string(
                name, prop_names, "Property name", not background, log_stdout
            )
            if name is None:
                log_stdout("Property name is required.")
                return
        except ValueError:
            log_stdout(f"Flag {name} not found in device {address}")

        if device.operating_flags.get(name):
            prop = device.operating_flags[name]
        else:
            prop = device.properties[name]

        try:
            if prop.value_type is bool:
                value = await self._ensure_bool(
                    value, "Property value", not background, log_stdout
                )
            else:
                value = await self._ensure_byte(
                    value, "Property value", not background, log_stdout
                )
            if value is None:
                log_stdout("Property value is required.")
                return
        except ValueError:
            log_stdout(f"Invalid value for {name}")
            return

        prop.new_value = value
        log_stdout(f"Changed property {prop.name} value to {value}")
        log_stdout("Use `write_config command to write changes to the device")

    async def do_write_config(self, address, log_stdout=None, background=False):
        """Write the device config.

        Usage:
            write_config <ADDRESS>|all
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=True,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        for device_address in addresses:
            device = devices[device_address]
            if device != devices.modem:
                if await device.async_write_op_flags() == ResponseStatus.SUCCESS:
                    log_stdout("Operating flags written")
                else:
                    log_stdout("Operating flags write error")
                if await device.async_write_ext_properties() == ResponseStatus.SUCCESS:
                    log_stdout("Extended properties written")
                else:
                    log_stdout("Extended properties write error")

    async def do_set_kpl_toggle_mode(
        self, address, button, toggle_mode, log_stdout=None, background=False
    ):
        """Set the toggle mode of a KeypadLinc button.

        Usage:
            set_kpl_toggle_mode address button toggle_mode

        address: Address of a KeypadLinc device
        button: Button number to set (2-8 on an 8 button KPL, 3-6 on a 6 button KPL)
        toggle_mode: Button toggle mode (0: toggle on/off, 1: On only, 2: Off only)
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]
        if not hasattr(device, "async_set_toggle_mode"):
            log_stdout("Device is not a KeypadLinc")
            return

        try:
            button = await self._ensure_int(
                button,
                device.groups.keys(),
                "Button",
                not background,
                log_stdout,
            )
            if button is None:
                log_stdout("A button number is required")
                return
        except ValueError:
            log_stdout("Invalid button number")
            return

        try:
            toggle_mode = await self._ensure_int(
                toggle_mode,
                [0, 1, 2],
                "Toggle mode",
                not background,
                log_stdout,
            )
            if toggle_mode is None:
                log_stdout("A toggle mode is required")
                return
        except ValueError:
            log_stdout("Invalid toggle mode")

        return await device.async_set_toggle_mode(
            button=button, toggle_mode=toggle_mode
        )

    async def do_set_kpl_radio_buttons(
        self, address, button1, button2, *args, log_stdout=None, background=False
    ):
        """Set the toggle mode of a KeypadLinc button.

        Usage:
            set_kpl_radio_buttons address button1 button2 [button3 button4 button5 button6 button7]

        address: Address of a KeypadLinc device
        button: Button number to set (2-8 on an 8 button KPL, 3-6 on a 6 button KPL)
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return
        device = devices[address]
        if not hasattr(device, "async_set_radio_buttons"):
            log_stdout("Device is not a KeypadLinc")
            return

        input_buttons = [button1, button2, *args]
        ask_values = button2 is None and not background
        buttons = []
        for button in range(0, len(device.groups)):
            try:
                input_button = input_buttons[button]
            except IndexError:
                input_button = None
            try:
                button = await self._ensure_int(
                    value=input_button,
                    values=device.groups.keys(),
                    name="Button",
                    ask_value=ask_values,
                    log_stdout=log_stdout,
                )
            except ValueError:
                if input_button is None and len(buttons) >= 2:
                    break
                log_stdout("Invalid button number")
                return
            if button is None and len(buttons) < 2:
                log_stdout("At least two buttons are required")
                return
            if button is None:
                break
            buttons.append(button)

        await device.async_set_radio_buttons(buttons=buttons)

    async def do_set_iolinc_mode(
        self, address, latching_mode, log_stdout=None, background=False
    ):
        """Set the relay mode of an IO Linc device.

        Usage:
            set_iolinc_mode address latching_mode

        address: IOLinc device address
        latching_mode: Trigger or Momentary A, B or C
            0 - Latching (Continuous)
            1 - Momentary A
            2 - Momentary B
            3 - Momentary C

        for more information see the IOLinc user manual.
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]
        if not hasattr(device, "async_set_relay_mode"):
            log_stdout("Device is not an IOLinc device")
            return

        try:
            latching_mode = await self._ensure_int(
                latching_mode, [0, 1, 2, 3], "Latching mode", not background, log_stdout
            )
            if latching_mode is None:
                log_stdout("Value for latching mode is required")
                return
        except ValueError:
            log_stdout("Invalid value for latching mode")
            return
        await device.async_set_relay_mode(latching_mode=RelayMode(latching_mode))

    async def do_set_iolinc_delay(
        self, address, seconds, log_stdout=None, background=False
    ):
        """Set the delay of an IO Linc device when in momentary mode.

        Usage:
            set_iolinc_delay address seconds

        address: IOLinc device address
        seconds: Integer number of seconds to delay before the relay is turned off

        For more information see the IOLinc user manual.
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]
        if not hasattr(device, "async_set_momentary_delay"):
            log_stdout("Device is not an IOLinc device")
            return

        try:
            seconds = await self._ensure_byte(
                seconds, "Delay (seconds)", not background, log_stdout
            )
            if seconds is None:
                log_stdout("Value for seconds is required")
        except ValueError:
            log_stdout("Invalid number for seconds")
            return

        await device.async_set_momentary_delay(seconds=seconds)

    async def do_get_engine_version(self, address, log_stdout=None, background=False):
        """Get the engine version of the device.

        Usage:
            get_engine_version <ADDRESS>|all
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=True,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        for device_address in addresses:
            device = devices[device_address]
            if not device == devices.modem and device.cat != 0x03:
                await device.async_get_engine_version()



"""I/O Sensor port configutation flags."""
from ..utils import bit_is_set, set_bit


class IOPortConfigFlags:
    """I/O Sensor port configuration flags."""

    def __init__(
        self,
        analog_input_mode: int = 0x00,
        send_on_sensor_alarm: bool = False,
        send_on_input_port_change: bool = False,
        enable_1_wire_port: bool = False,
        enable_all_link_aliasing: bool = False,
        send_on_outut_port_change: bool = False,
        enable_output_timers: bool = False,
    ):
        """Init the IOPortConfigFlags class."""
        # Analog Input Mode:
        #   0x00 Analog Input not used
        #   0x01 Analog Input used, convert upon command
        #   0x10 Analog Input used, convert at fixed interval
        #   0x11 Unused
        self._analog_input_mode = analog_input_mode
        self._send_on_sensor_alarm = send_on_sensor_alarm
        self._send_on_input_port_change = send_on_input_port_change
        self._enable_1_wire_port = enable_1_wire_port
        self._enable_all_link_aliasing = enable_all_link_aliasing
        self._send_on_output_port_change = send_on_outut_port_change
        self._enable_output_timers = enable_output_timers

    def __bytes__(self):
        """Return the byte representation of the flags."""
        return bytes(bytearray([int(self)]))

    def __int__(self):
        """Return the integer representation of the flags."""
        flags = self._analog_input_mode
        flags = set_bit(flags, 2, self._send_on_sensor_alarm)
        flags = set_bit(flags, 3, self._send_on_input_port_change)
        flags = set_bit(flags, 4, self._enable_1_wire_port)
        flags = set_bit(flags, 5, self._enable_all_link_aliasing)
        flags = set_bit(flags, 6, self._send_on_output_port_change)
        flags = set_bit(flags, 7, self._enable_output_timers)
        return flags

    @property
    def analog_input_mode(self):
        """Return the analog_input_mode flag."""
        return self._analog_input_mode

    @property
    def send_on_sensor_alarm(self):
        """Return the send_on_sensor_alarm flag."""
        return self._send_on_sensor_alarm

    @property
    def send_on_input_port_change(self):
        """Return the send_on_input_port_change flag."""
        return self._send_on_input_port_change

    @property
    def enable_1_wire_port(self):
        """Return the enable_1_wire_port flag."""
        return self._enable_1_wire_port

    @property
    def enable_all_link_aliasing(self):
        """Return the enable_all_link_aliasing flag."""
        return self._enable_all_link_aliasing

    @property
    def send_on_output_port_change(self):
        """Return the send_on_output_port_change flag."""
        return self._send_on_output_port_change

    @property
    def enable_output_timers(self):
        """Return the enable_output_timers flag."""
        return self._enable_output_timers

    @classmethod
    def create_from_byte(cls, value):
        """Create a IOPortConfigFlags from a byte value."""
        analog_input_mode = value & 0x03
        send_on_sensor_alarm = bit_is_set(value, 2)
        send_on_input_port_change = bit_is_set(value, 2)
        enable_1_wire_port = bit_is_set(value, 2)
        enable_all_link_aliasing = bit_is_set(value, 2)
        send_on_outut_port_change = bit_is_set(value, 2)
        enable_output_timers = bit_is_set(value, 2)
        return cls(
            analog_input_mode,
            send_on_sensor_alarm,
            send_on_input_port_change,
            enable_1_wire_port,
            enable_all_link_aliasing,
            send_on_outut_port_change,
            enable_output_timers,
        )



"""Handler for exceptions thrown by pubsub listeners."""
from . import pub


class GeneralException(Exception):
    """General exception handler."""


class ListenerExceptionHandler(pub.IListenerExcHandler):
    """Handle exceptions from listeners of pubsub topics."""

    def __call__(self, listenerID, topicObj):
        """Call the exception handler.

        This forces the pub.ExcHandlerError to be raised.
        """
        raise GeneralException



"""Utility methods."""

import asyncio
from collections.abc import Iterable
from enum import Enum, IntEnum
from functools import partial, wraps
from inspect import isawaitable, iscoroutinefunction
import logging
import traceback

from . import pub
from .address import Address
from .commands import commands
from .constants import (
    HC_LOOKUP,
    RAMP_RATES,
    RAMP_RATES_SEC,
    UC_LOOKUP,
    FanSpeed,
    FanSpeedRange,
    MessageFlagType,
    ResponseStatus,
    ThermostatMode,
    X10Commands,
)
from .topics import STATUS_REQUEST

_LOGGER = logging.getLogger(__name__)
_LOGGER_TOPICS = logging.getLogger("pyinsteon.topics")


def housecode_to_byte(housecode: str) -> int:
    """Return the byte value of an X10 housecode."""
    return HC_LOOKUP.get(housecode.lower())


def unitcode_to_byte(unitcode: int) -> int:
    """Return the byte value of an X10 unitcode."""
    return UC_LOOKUP.get(unitcode)


def byte_to_housecode(bytecode: int) -> str:
    """Return an X10 housecode value from a byte value."""
    house_code = list(HC_LOOKUP.keys())[list(HC_LOOKUP.values()).index(bytecode)]
    return house_code.upper()


def byte_to_unitcode(bytecode: int) -> int:
    """Return an X10 unitcode value from a byte value."""
    return list(UC_LOOKUP.keys())[list(UC_LOOKUP.values()).index(bytecode)]


def byte_to_command(bytecode: int) -> int:
    """Return an X10 unitcode value from a byte value."""
    return X10Commands(bytecode)


def parse_x10(raw_x10):
    """Convert an X10 message to a dictionary."""
    housecode, uc_or_cmd = raw_x10_to_bytes(raw_x10)
    return byte_to_housecode(housecode), uc_or_cmd


def byte_to_int(bytecode: bytes) -> int:
    """Return an int from a byte string."""
    return int.from_bytes(bytecode, byteorder="big")


def raw_x10_to_bytes(raw_x10: int) -> int:
    """Return the byte value of a raw X10 command."""
    house_code_byte = raw_x10 >> 4
    uc_or_cmd_byte = raw_x10 & 0x0F
    return house_code_byte, uc_or_cmd_byte


def bit_is_set(bitmask: int, bit: int) -> bool:
    """Return True if a specific bit is set in a bitmask.

    Uses the low bit is 1 and the high bit is 8.
    """
    return bool(bitmask & (1 << bit))


def set_bit(data: int, bit: int, is_on: bool) -> int:
    """Set the value of a bit in a bitmask on or off.

    Uses the low bit is 0 and the high bit is 7.
    """
    if isinstance(data, bytes):
        bitmask = int.from_bytes(data, byteorder="big")
    else:
        bitmask = data
    if is_on:
        return bitmask | (1 << bit)
    return bitmask & (0xFF & ~(1 << bit))


def vars_to_bytes(vals: Iterable) -> bytes:
    """Create a byte string from a set of values."""
    msg = bytearray()
    for val in vals:
        if val is None:
            pass
        elif isinstance(val, int):
            msg.extend(bytes([val]))
        else:
            msg.extend(bytes(val))
    return bytes(msg)


def vars_to_string(vals: Iterable) -> str:
    """Create a byte string from a set of values."""

    output = []
    for fld, val in vals:
        if val is None:
            pass
        elif isinstance(val, (Enum, IntEnum)):
            valstr = str(val)
        elif isinstance(val, int):
            valstr = f"0x{val:02x}"
        elif isinstance(val, bytes):
            valstr = f"0x{val.hex():s}"
        else:
            valstr = str(val)
        output.append(f"{fld}: {valstr}")
    return ", ".join(output)


def vars_to_repr(vals: Iterable) -> str:
    """Create a byte string from a set of values."""
    output = []
    for fld, val in vals:
        if val is None:
            pass
        elif isinstance(val, (Enum, IntEnum)):
            valstr = repr(val)
        elif isinstance(val, int):
            valstr = f"0x{val:02x}"
        elif isinstance(val, bytes):
            valstr = f"0x{val.hex():s}"
        else:
            valstr = repr(val)
        output.append(f"{fld}: {valstr}")
    return ", ".join(output)


def test_values_eq(val1, val2) -> bool:
    """Test if val1 eq val2 for template management."""
    if val1 is None or val2 is None:
        return True
    if val1 == val2:
        return True
    return False


def _include_address(prefix, topic, address, message_type):
    """Test if we should include the address in the topic build."""
    if address is None:
        return False

    if isinstance(message_type, str):
        message_type = getattr(MessageFlagType, str(message_type).upper())

    if prefix == "send" and message_type == MessageFlagType.DIRECT:
        return False

    if prefix == "handler" and message_type == MessageFlagType.DIRECT:
        return True

    if commands.get(topic) is None:
        return False

    return True


def _msg_group(topic, group, message_type):
    if topic == STATUS_REQUEST and group is not None:
        return group

    if commands.use_group(topic) and group is not None:
        return group if group else 1

    if group is None:
        return None

    if not commands.use_group(topic):
        return None

    if message_type in [MessageFlagType.ALL_LINK_CLEANUP]:
        return None

    return 1


def build_topic(topic, prefix=None, address=None, group=None, message_type=None):
    """Build a full topic from components."""
    full_topic = ""
    if prefix is not None:
        # Adding the . separator since there must be something after a prefix
        full_topic = f"{str(prefix).lower()}."

    if _include_address(prefix, topic, address, message_type):
        addr = address.id if isinstance(address, Address) else address
        full_topic = f"{full_topic}{addr}."
        group = _msg_group(topic, group, message_type)
        if group is not None:
            full_topic = f"{full_topic}{group}."

    full_topic = f"{full_topic}{topic}"
    if message_type is not None:
        full_topic = f"{full_topic}.{str(message_type).lower()}"
    return full_topic


def multiple_status(*args):
    """Return the proper status based on the worst case of all status responses."""
    worst_response = 1
    for response in args:
        if response is None:
            continue
        if int(response) == 0 or worst_response == 0:
            worst_response = 0
        elif int(response) > worst_response:
            worst_response = int(response)
    return ResponseStatus(worst_response)


def ramp_rate_to_seconds(ramp_rate: int):
    """Return the seconds associated with a ramp rate."""
    if int(ramp_rate) not in range(0, 32):
        raise ValueError("Ramp rate must be between 0x00 and 0x1f (31)")

    return RAMP_RATES[int(ramp_rate)]


def _abs_diff(list_value, test_value):
    """Return the absolute difference between two numbers."""
    return abs(list_value - test_value)


def seconds_to_ramp_rate(seconds: float):
    """Return the ramp rate asscociated with a number of seconds."""

    if seconds > 480 or seconds < 0.1:
        raise ValueError(
            "Ramp rate cannot be less than 0.1 seconds or more than 480 seconds (8 minutes)"
        )

    rr_sec_list = list(map(float, RAMP_RATES_SEC.keys()))
    abs_diff = partial(_abs_diff, test_value=seconds)
    rr_sec = min(rr_sec_list, key=abs_diff)

    return RAMP_RATES_SEC[rr_sec]


def log_error(msg, ex, topic=None, kwargs=None):
    """Print an error message when a topic cannot be distributed."""
    _LOGGER.error("An issue occured distributing the following message")
    _LOGGER.error("MSG: %s", msg)
    _LOGGER.error("Topic: %s data: %s", topic, kwargs)
    _LOGGER.error("Error: %s", str(ex))
    _LOGGER_TOPICS.debug(traceback.format_exc())
    if topic is not None:
        topic_mgr = pub.getDefaultTopicMgr()
        topic = topic_mgr.getTopic(topic, okIfNone=True)
        if topic:
            for subscriber in topic.getListeners():
                _LOGGER.error("    Subscriber: %s", subscriber)


def to_celsius(fahrenheit):
    """Convert fahrenheit to celsius."""
    return int(round((fahrenheit - 32) * 5 / 9, 0))


def to_fahrenheit(celsius):
    """Convert celsius to fahrenheit."""
    return int(round(celsius * 9 / 5 + 32, 0))


def calc_thermostat_temp(high_byte, low_byte):
    """Calculate the temperature."""
    return round((low_byte | (high_byte << 8)) * 0.1, 1)


def calc_thermostat_mode(thermostat_mode_byte, sys_mode_map=None, sys_low=True):
    """Calculate the system and fan mode."""
    if sys_mode_map is None:
        sys_mode_map = {
            int(ThermostatMode.OFF): ThermostatMode.OFF,
            int(ThermostatMode.AUTO): ThermostatMode.AUTO,
            int(ThermostatMode.HEAT): ThermostatMode.HEAT,
            int(ThermostatMode.COOL): ThermostatMode.COOL,
        }

    mode1 = thermostat_mode_byte & 0x0F
    mode2 = thermostat_mode_byte >> 4
    system_mode, fan_mode = (mode1, mode2) if sys_low else (mode2, mode1)
    if fan_mode in (0, 4):
        fan_mode = ThermostatMode.FAN_AUTO
    else:
        fan_mode = ThermostatMode.FAN_ALWAYS_ON
    sys_mode = sys_mode_map.get(system_mode)
    if sys_mode is None:
        sys_mode = ThermostatMode.AUTO
    return sys_mode, fan_mode


def publish_topic(topic, logger=None, **kwargs):
    """Publish a topic and log errors."""
    # Send log message as caller not utils.
    if logger is None:
        logger = logging.getLogger(__name__)
    try:
        pub_topic = pub.getDefaultTopicMgr().getTopic(topic, okIfNone=True)
        if pub_topic:
            for listener in pub_topic.listeners:
                if listener.isDead():
                    pub_topic.unsubscribe(listener)
        pub.sendMessage(topic, **kwargs)
    except pub.ExcHandlerError as exc:
        logger.error("pubsub ExcHandlerError")
        logger.error("Error processing topic: %s", topic)
        logger.error("Error in topic listner: %s", exc.badExcListenerID)
    except pub.SenderMissingReqdMsgDataError as exc:
        logger.error("SenderMissingReqdMsgDataError")
        logger.error(str(exc))
        for listner in pub.getDefaultTopicMgr().getTopic(topic).getListeners():
            logger.error("Topic listener: %s", listner)
    except pub.SenderUnknownMsgDataError as exc:
        logger.error("SenderUnknownMsgDataError")
        logger.error(str(exc))
        for listner in pub.getDefaultTopicMgr().getTopic(topic).getListeners():
            logger.error("Topic listener: %s", listner)


async_listeners = {}


def get_async_listener(listener):
    """Get the async version of a listener."""

    def setup_async_listener(listener):
        @wraps(listener)
        def _wrapper(*args, **kwargs):
            return asyncio.create_task(listener(*args, **kwargs))

        return _wrapper

    async_listener = async_listeners.get(listener)
    if async_listener:
        return async_listener
    async_listener = setup_async_listener(listener)
    async_listeners[listener] = async_listener
    return async_listener


def remove_async_listener(listener):
    """Remove an async listener."""
    async_listeners.pop(listener)


def subscribe_topic(listener, topic_name, logger=None):
    """Subscribe a listener to a topic and log errors."""

    topic_mgr = pub.getDefaultTopicMgr()
    topic = topic_mgr.getOrCreateTopic(topic_name)
    if topic.hasListener(listener):
        return
    if logger is None:
        logger = logging.getLogger(__name__)
    try:
        if iscoroutinefunction(listener) or isawaitable(listener):
            async_listener = get_async_listener(listener)
            if topic.hasListener(async_listener):
                return
            topic.subscribe(async_listener)
        else:
            topic.subscribe(listener)
    except pub.ListenerMismatchError as exc:
        logger.error("ListenerMismatchError")
        logger.error("args: %s", exc.args)
        logger.error("msg: %s", exc.msg)
        logger.error("module: %s", exc.module)
        logger.error("idStr: %s", exc.idStr)
        for topic_listner in pub.getDefaultTopicMgr().getTopic(topic).getListeners():
            logger.error("Topic listener: %s", topic_listner)


def unsubscribe_topic(listener, topic_name):
    """Unsubscribe a listener to a topic and log errors."""
    listener = async_listeners.get(listener, listener)
    topic_mgr = pub.getDefaultTopicMgr()
    topic = topic_mgr.getOrCreateTopic(topic_name)
    if pub.isSubscribed(listener, topicName=topic.name):
        pub.unsubscribe(listener, topic_name)


def set_fan_speed(on_level):
    """Map a value to a fan speed."""
    if on_level in FanSpeedRange.OFF.value:
        return FanSpeed.OFF
    if on_level in FanSpeedRange.LOW.value:
        return FanSpeed.LOW
    if on_level in FanSpeedRange.MEDIUM.value:
        return FanSpeed.MEDIUM
    return FanSpeed.HIGH



"""Handle All-Link Cleanup Status Reports."""
from ..address import Address
from ..topics import ALL_LINK_CLEANUP_FAILURE_REPORT
from . import inbound_handler
from .inbound_base import InboundHandlerBase


class AllLinkCleanupFailureReport(InboundHandlerBase):
    """Handle All-Link Cleanup Failure Reports."""

    def __init__(self):
        """Init AllLinkCleanupFailureReport class."""
        super().__init__(topic=ALL_LINK_CLEANUP_FAILURE_REPORT)

    @inbound_handler
    def handle_inbound(self, error: int, group: int, target: Address):
        """Handle the All-Link Failure Report inbound message."""
        self._call_subscribers(success=False, group=group, target=target)



"""Operating flag byte for a property read/write process.

For use with `ExtendedPropertyReadManager` and `ExtendedPropertyWriteManager`.

- Receives a list of operating flags and the associated bit.
- Stores the value of a property byte just like any other property.
- Updates the associated operating flags with the correct value based on the bit mapping.
- For bits that do not have an operating flag, the underlying `_value` property will determine the bit value.

"""
from typing import Dict

from ..constants import PropertyType
from ..utils import bit_is_set, set_bit
from .device_flag import DeviceFlagBase
from .operating_flag import OperatingFlag


def _value(prop: OperatingFlag):
    return prop.new_value if prop.is_dirty else prop.value


class OpFlagPropertyByte(DeviceFlagBase):
    """Operating flag byte for a property read/write process.

    For use with `ExtendedPropertyReadManager` and `ExtendedPropertyWriteManager`.
    """

    def __init__(
        self, address, name, flags: Dict[int, OperatingFlag], is_read_only=False
    ):
        """Init the DerivedProperty class."""
        super().__init__(
            address,
            "derived_property",
            name,
            int,
            is_reversed=False,
            is_read_only=is_read_only,
        )
        self._flags: Dict[int, OperatingFlag] = flags
        self._prop_type = PropertyType.HIDDEN

    @property
    def is_dirty(self):
        """Return if any of the operating flags are dirty."""
        for _, prop in self._flags.items():
            if prop.is_dirty:
                return True
        return False

    @property
    def is_loaded(self):
        """Return if the Operating flag has been loaded."""
        for _, prop in self._flags.items():
            if not prop.is_loaded:
                return False
        return True

    @property
    def value(self):
        """Return the value of the flag."""
        byte_value = self._value if self._value is not None else 0x00
        for bit, prop in self._flags.items():
            if prop.value is None:
                return None
            byte_value = set_bit(byte_value, bit, prop.value)
        return byte_value

    @property
    def new_value(self):
        """Return the new value of the flag."""
        byte_value = self.value
        for bit, prop in self._flags.items():
            if prop.value is None:
                return None
            value = _value(prop)
            byte_value = set_bit(byte_value, bit, value)
        return byte_value

    @new_value.setter
    def new_value(self, value):
        """Set the new value of the flag.

        This is the primary method to set the value of the flag.
        It sets the `new_value` property and the `is_dirty` property.
        """
        for bit, prop in self._flags.items():
            new_val = bit_is_set(value, bit) if value is not None else None
            prop.new_value = new_val

    def set_value(self, value):
        """Load the flag from the device value.

        Only use this method to update the value of the flag from the value
        of the device.

        This method updates the `is_loaded` property and clears the `new value` and
        `is_dirty` properties.
        """
        super().set_value(value=value)
        for bit, prop in self._flags.items():
            if value is None:
                new_val = None
            else:
                new_val = bit_is_set(value, bit)
            prop.set_value(new_val)



"""Operating flag or Extended Property for all device types."""
from ..address import Address
from ..constants import PropertyType
from ..subscriber_base import SubscriberBase


class DeviceFlagBase(SubscriberBase):
    """Operating flag or Extended Property."""

    def __init__(
        self,
        address,
        topic_prefix,
        name,
        value_type: type,
        is_reversed=False,
        is_read_only=False,
        prop_type=PropertyType.STANDARD,
    ):
        """Init the DeviceFlag class."""
        self._address = Address(address)
        topic = f"{self._address.id}.{topic_prefix}.{name}"
        super().__init__(topic)
        self._name = name
        self._value = None
        self._value_type = value_type
        self._new_value = None
        self._is_loaded = False
        self._reversed = is_reversed
        self._read_only = is_read_only
        self._prop_type = PropertyType(prop_type)

    @property
    def name(self):
        """Return the flag name."""
        return self._name

    @property
    def value(self):
        """Return the value of the flag."""
        return self._value

    @property
    def new_value(self):
        """Return the new value of the flag."""
        return self._new_value

    @new_value.setter
    def new_value(self, value):
        """Set the new value of the flag.

        This is the primary method to set the value of the flag.
        It sets the `new_value` property and the `is_dirty` property.
        """
        if self._read_only:
            return

        if self._value_type is bool and self._reversed:
            value = not value

        if value != self._value and value is not None:
            self._new_value = self._value_type(value)
        else:
            self._new_value = None

    @property
    def value_type(self):
        """Return the property type."""
        return self._value_type

    @property
    def is_dirty(self):
        """Return if the Operating flag has been changed."""
        return self._new_value is not None

    @property
    def is_read_only(self):
        """Return the read only flag."""
        return self._read_only

    @is_read_only.setter
    def is_read_only(self, value: bool) -> None:
        """Set the read only flag of the property."""
        self._read_only = bool(value)

    @property
    def is_loaded(self):
        """Return if the Operating flag has been loaded."""
        return self._is_loaded

    @property
    def is_reversed(self):
        """Return if the device flag is reverse of this value."""
        return self._reversed

    @property
    def property_type(self):
        """Return the property type (Standard, Advanced, Derived or Hidden)."""
        return self._prop_type

    @property_type.setter
    def property_type(self, value: PropertyType):
        """Set the propoerty type for this property."""
        self._prop_type = PropertyType(value)

    def set_value(self, value):
        """Load the flag from the device value.

        Only use this method to update the value of the flag from the value
        of the device.

        This method updates the `is_loaded` property and clears the `new value` and
        `is_dirty` properties.
        """
        if self._value_type is bool and self._reversed:
            value = not value
        update_value = (
            self._value_type(value) if value is not None else self._value_type(0)
        )
        self._new_value = None
        self._is_loaded = True
        is_changed = update_value != self._value
        self._value = update_value
        if is_changed:
            self._call_subscribers(name=self._name, value=self._value)



"""Base class for all group value entities."""

from abc import ABC, abstractmethod

from ..address import Address
from ..subscriber_base import SubscriberBase


class GroupBase(SubscriberBase):
    """Device group base class."""

    __meta__ = ABC

    def __init__(
        self,
        name: str,
        address: Address,
        group=0,
        default=None,
        value_type: type = int,
        status_type: int = 0,
    ):
        """Init the StateBase class."""
        self._address = address
        topic = f"state_{self._address.id}_{name}_{group}"
        super().__init__(subscriber_topic=topic)
        self._name = name
        self._group = group
        self._value = int(default) if default is not None else None
        self._type = value_type
        self._is_dimmable: bool = False
        self._status_type = status_type

    @property
    def is_dimmable(self) -> bool:
        """Return if the state is dimmable.

        If true, the state can support values 0 - 255.
        If false, the state only supports values 0 and 255.
        """
        return self._is_dimmable

    @is_dimmable.setter
    def is_dimmable(self, value: bool) -> None:
        """Set the dimmable property of the state."""
        self._is_dimmable = bool(value)

    @property
    def name(self):
        """Return the name of the state."""
        return self._name

    @property
    def group(self):
        """Return the group number."""
        return self._group

    @property
    def value(self):
        """Return the value of the state."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the state."""
        try:
            value = self._type(value) if value is not None else None
            if self._value == value:
                return
            self._value = value
        except TypeError as ex:
            raise TypeError(
                f"Error setting value of State {self._name}: Must be of type {self._type.__name__}"
            ) from ex
        self._call_subscribers(
            name=self._name,
            address=self._address.id,
            value=self._value,
            group=self._group,
        )

    @property
    def status_type(self):
        """Return the status type to get the group state."""
        return self._status_type

    @abstractmethod
    def set_value(self, **kwargs):
        """Set the value of the state from a Handler."""
        raise NotImplementedError



"""Wrapper for serial.aio transport."""
import asyncio
import logging

import serial
from serial_asyncio import SerialTransport

_LOGGER = logging.getLogger(__name__)


async def async_connect_serial(device, protocol):
    """Connect to the PowerLine Modem via serial port.

    Parameters:
        port – Device name.
        protocol - Insteon Modem Protocol instance.

    """
    loop = asyncio.get_event_loop()
    try:
        ser = serial.serial_for_url(url=device, baudrate=19200)
        transport = SerialTransportInsteon(loop, protocol, ser, device=device)
    except OSError as ex:
        _LOGGER.warning("Unable to connect to %s: %s", device, ex)
        transport = None
    return transport


async def async_connect_socket(host, protocol, port=None):
    """Connect to the Hub Version 1 via TCP Socket."""

    port = 9761 if not port else port
    loop = asyncio.get_event_loop()
    url = f"socket://{host}:{port}"
    try:
        ser = serial.serial_for_url(url=url, baudrate=19200)
        transport = SerialTransportInsteon(loop, protocol, ser, device=url)
    except OSError as ex:
        _LOGGER.warning("Unable to connect to %s: %s", url, ex)
        transport = None
    return transport


class SerialTransportInsteon(SerialTransport):
    """Wrapper for serial_asyncio.SerialTransport."""

    def __init__(self, loop, protocol, serial_instance, device):
        """Init the SerialTransportInsteon class."""
        super().__init__(loop=loop, protocol=protocol, serial_instance=serial_instance)
        self._device = device

    @property
    def connected(self):
        """Return True if the transport is connected to the serial device."""
        return not self._closing

    @property
    def write_wait(self):
        """Return the time to wait between writes."""
        return 0.8

    def write(self, data):
        """Override SerialTransport write method."""
        msg_bytes = bytes(data)
        super().write(msg_bytes)

    async def async_write(self, data):
        """Asyncronous write method."""
        self.write(data)



"""Insteon Modem ALDB Read Manager."""
import asyncio
import logging

import async_timeout

from .. import pub
from ..address import Address
from ..aldb.aldb_record import ALDBRecord
from ..constants import ManageAllLinkRecordAction, ResponseStatus
from ..handlers.all_link_record_response import AllLinkRecordResponseHandler
from ..handlers.get_first_all_link_record import GetFirstAllLinkRecordHandler
from ..handlers.get_next_all_link_record import GetNextAllLinkRecordHandler
from ..handlers.manage_all_link_record import ManageAllLinkRecordCommand
from ..handlers.read_eeprom import ReadEepromHandler
from ..handlers.read_eeprom_response import ReadEepromResponseHandler
from ..topics import ALL_LINK_RECORD_RESPONSE

_LOGGER = logging.getLogger(__name__)
TIMEOUT = 2


class ImReadManager:
    """Insteon Modem ALDB Read Manager."""

    def __init__(self, aldb):
        """Init the ImReadManager class."""
        self._aldb = aldb
        self._get_first_handler = GetFirstAllLinkRecordHandler()
        self._get_next_handler = GetNextAllLinkRecordHandler()
        self._find_cmd = ManageAllLinkRecordCommand()
        mgr = pub.getDefaultTopicMgr()
        topic = mgr.getTopic(ALL_LINK_RECORD_RESPONSE, okIfNone=True)
        if not topic:
            self._receive_record_handler = AllLinkRecordResponseHandler()
            self._receive_record_handler.subscribe(self._receive_record)
            self._receive_eeprom_record_handler = ReadEepromResponseHandler()
            self._receive_eeprom_record_handler.subscribe(self._receive_eeprom_record)
        self._retries = 0
        self._load_lock = asyncio.Lock()
        self._record_queue = asyncio.Queue()

    async def async_load_standard(self):
        """Load the Insteon Modem ALDB using standard method."""
        self._clear_read_queue()
        record = await self.async_get_first_all_link_record()
        if record:
            yield record
        else:
            return
        async for record in self.async_get_next_all_link_records():
            yield record

    async def async_load_eeprom(self):
        """Load the Insteon modem ALDB using EEPROM reads."""
        self._clear_read_queue()
        next_mem_addr = self._aldb.first_mem_addr
        record = await self.async_read_record(next_mem_addr)
        if record:
            yield record
        else:
            return
        for record in await self.async_read_record(next_mem_addr):
            yield record
            if record.is_high_water_mark:
                return
            next_mem_addr -= 8

    async def async_get_first_all_link_record(self):
        """Get the first All-Link database record."""
        self._clear_read_queue()
        retries = 3
        response = None
        while retries and response != ResponseStatus.SUCCESS:
            response = await self._get_first_handler.async_send()
            _LOGGER.debug("Get First Response: %s", response)
            retries -= 1

        if response == ResponseStatus.SUCCESS:
            try:
                async with async_timeout.timeout(TIMEOUT):
                    return await self._record_queue.get()
            except asyncio.TimeoutError:
                pass
        return None

    async def async_get_next_all_link_records(self):
        """Get the next All-Link database record."""
        self._clear_read_queue()
        retries = 3
        result = None
        while retries:
            result = await self._get_next_handler.async_send()
            retries -= 1

            if result == ResponseStatus.SUCCESS:
                try:
                    async with async_timeout.timeout(TIMEOUT):
                        yield await self._record_queue.get()
                except asyncio.TimeoutError:
                    pass
                else:
                    retries = 3
        return

    async def async_find(self, target, group):
        """Find one or more ALDB records."""
        self._clear_read_queue()
        record = await self._find_first_record(target=target, group=group)
        if record:
            yield record
        else:
            return
        async for record in self._find_next_records(target=target, group=group):
            yield record

    async def async_read_record(self, mem_addr: int) -> ALDBRecord:
        """Read from EEPROM."""
        self._clear_read_queue()
        retries = 20
        cmd = ReadEepromHandler()
        response = None
        while retries and response != ResponseStatus.SUCCESS:
            response = await cmd.async_send(mem_addr=mem_addr)
            retries -= 1
            if response == ResponseStatus.SUCCESS:
                try:
                    async with async_timeout.timeout(TIMEOUT):
                        return await self._record_queue.get()
                except asyncio.TimeoutError:
                    return None
            await asyncio.sleep(0.5)

    async def async_confirm_eeprom_read(self) -> int:
        """Confirm the first memory address is readable with an EEPROM read.

        Utilizes "get first aldb record" and "read eeprom" read methods to confirm
        the EEPROM method returns the same record as the standard modem read method.

        """
        mem_addrs = [0x1FFF, 0x3FFF]
        if self._aldb.first_mem_addr in mem_addrs:
            mem_addrs.remove(self._aldb.first_mem_addr)
        # Make the default memory address the first one we check
        mem_addrs.insert(0, self._aldb.first_mem_addr)

        self._clear_read_queue()
        async with self._load_lock:
            first_record = await self.async_get_first_all_link_record()
            eeprom_record = None
            for mem_addr in mem_addrs:
                eeprom_record = await self.async_read_record(mem_addr)
                if eeprom_record:
                    break

            if first_record is None and eeprom_record.is_high_water_mark:
                return eeprom_record.mem_addr  # ALDB is empty

            if first_record.is_exact_match(eeprom_record):
                return eeprom_record.mem_addr

            return 0

    async def _find_first_record(self, target, group):
        """Find the records in the background."""
        retries = 3
        response = None
        while retries and response != ResponseStatus.SUCCESS:
            response = await self._find_cmd.async_send(
                action=ManageAllLinkRecordAction.FIND_FIRST,
                controller=False,
                group=group,
                target=Address(target),
                data1=0,
                data2=0,
                data3=0,
                in_use=True,
                high_water_mark=False,
                bit5=0,
                bit4=0,
            )
            if response == ResponseStatus.SUCCESS:
                try:
                    async with async_timeout.timeout(TIMEOUT):
                        return await self._record_queue.get()
                except asyncio.TimeoutError:
                    pass
            retries -= 1

    def _receive_record(
        self,
        in_use: bool,
        high_water_mark: bool,
        controller: bool,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
        bit5: bool,
        bit4: bool,
    ):
        """Receive a record and load into the ALDB."""
        record = ALDBRecord(
            memory=0,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
        self._record_queue.put_nowait(record)

    def _receive_eeprom_record(
        self,
        mem_addr: int,
        in_use: bool,
        high_water_mark: bool,
        controller: bool,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
        bit5: bool,
        bit4: bool,
    ):
        """Receive an ALDB record from an EEPROM read."""
        record = ALDBRecord(
            memory=mem_addr,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
        self._record_queue.put_nowait(record)

    async def _find_next_records(self, target, group):
        """Find the next ALDB records."""
        response = ResponseStatus.FAILURE
        retries = 3
        unsent = 50
        while retries and unsent:
            response = await self._find_cmd.async_send(
                action=ManageAllLinkRecordAction.FIND_NEXT,
                controller=False,
                group=group,
                target=Address(target),
                data1=0,
                data2=0,
                data3=0,
                in_use=True,
                high_water_mark=False,
                bit5=0,
                bit4=0,
            )
            if response == ResponseStatus.SUCCESS:
                try:
                    async with async_timeout.timeout(TIMEOUT):
                        yield await self._record_queue.get()
                except asyncio.TimeoutError:
                    retries -= 1
                else:
                    retries = 3
                    unsent = 50
            elif response == ResponseStatus.UNSENT:
                unsent -= 1
            else:
                retries -= 1

        yield None

    def _clear_read_queue(self):
        """Clear the read queue of old records."""

        while not self._record_queue.empty():
            self._record_queue.get_nowait()



"""Methods to convert to and from topics."""
from ..constants import MessageFlagType


def topic_to_message_type(topic):
    """Return MessageFlagType from the topic."""
    subtopics = topic.name.split(".")
    flag = "direct" if len(subtopics) < 3 else subtopics[2]
    for msg_type in MessageFlagType:
        if flag.lower() == str(msg_type):
            return msg_type
    return MessageFlagType(0)


def topic_to_message_handler(topic, register_list):
    """Register handler to topic."""

    def register(func):
        register_list[f"send.{topic}"] = func
        return func

    return register


def topic_to_command_handler(topic, register_list):
    """Register handler to topic."""

    def register(func):
        register_list[f"send.{topic}"] = func
        return func

    return register



"""Python module for controlling Insteon devices."""
import asyncio
import logging

from pubsub import pub

from .address import Address
from .handlers.from_device.x10_received import X10Received
from .listener_exception_handler import ListenerExceptionHandler
from .managers.device_link_manager import DeviceLinkManager
from .managers.device_manager import DeviceManager
from .managers.link_manager.default_links import async_add_default_links

# pylint: disable=unused-import
from .managers.x10_manager import async_x10_all_lights_off  # noqa: F401
from .managers.x10_manager import async_x10_all_lights_on  # noqa: F401
from .managers.x10_manager import async_x10_all_units_off  # noqa: F401
from .protocol import async_modem_connect
from .topics import ADD_DEFAULT_LINKS
from .utils import subscribe_topic

_LOGGER_TOPICS = logging.getLogger("pyinsteon.topics")
X10_RECEIVED_HANDLER = X10Received()

devices = DeviceManager()
link_manager = DeviceLinkManager(devices)
subscribe_topic(async_add_default_links, ADD_DEFAULT_LINKS)


async def async_connect(
    device=None,
    host=None,
    port=None,
    username=None,
    password=None,
    hub_version=2,
    **kwargs,
):
    """Connect to the Insteon Modem.

    Parameters:
        device: PLM serial / USB device name
        host: Hub hostname or IP address
        port: Hub TCP port
        username: Hub v2 username
        password: Hub v2 password
        hub_version: Hub version (default=2)

    Returns an Insteon Modem (PLM or Hub).

    """
    try:
        modem = await async_modem_connect(
            device=device,
            host=host,
            port=port,
            username=username,
            password=password,
            hub_version=hub_version,
            **kwargs,
        )
    except ConnectionError as err:
        raise ConnectionError from err
    devices.modem = modem
    devices.id_manager.start()
    await devices.modem.async_get_configuration()
    return devices


async def async_close():
    """Close the connection and stop all tasks."""
    await devices.modem.async_close()
    for addr in devices:
        if devices[addr].is_battery:
            devices[addr].close()
    devices.id_manager.close()
    await asyncio.sleep(0.1)


def set_exception_handler():
    """Set the pubsub exception handler.

    This should only be run for debugging purposes.
    """
    pub.setListenerExcHandler(ListenerExceptionHandler())


def _get_device_in_topic(topic: pub.Topic, **kwargs):
    """Return a list of devices referenced in a topic."""
    addresses = []
    prefix = ["ack", "nak", "handler"]
    topic_tuple = topic.getNameTuple()
    try:
        if topic_tuple[0] == "send":
            pass
        elif topic_tuple[0] in prefix:
            addresses.append(Address(topic_tuple[1]))
        else:
            addresses.append(Address(topic_tuple[0]))
    except ValueError:
        pass

    for key, value in kwargs.items():
        if key in ["address", "target"]:
            try:
                addr = Address(value)
            except ValueError:
                continue
            else:
                if addr not in addresses and not str(addr).startswith("000"):
                    addresses.append(addr)

    return addresses


def _log_all_topics(topic=pub.AUTO_TOPIC, **kwargs):
    """Log all topics from pyinsteon."""
    _LOGGER_TOPICS.debug("Topic: %s data: %s", topic.name, kwargs)
    if _LOGGER_TOPICS.level > logging.DEBUG:
        addresses = _get_device_in_topic(topic, **kwargs)
        for addr in addresses:
            logger = logging.getLogger(f"pyinsteon.{addr.id}")
            logger.debug("Topic: %s data: %s", topic.name, kwargs)


pub.subscribe(_log_all_topics, pub.ALL_TOPICS)



"""Mock transport for test purposes."""
import asyncio
from binascii import Error, Incomplete, unhexlify
import logging

from ...address import Address
from ...constants import AckNak, MessageFlagType, MessageId
from ...data_types.message_flags import MessageFlags
from .mock_reader import MockReader

_LOGGER = logging.getLogger(__name__)
READ_WAIT = 0.5
SESSION_RETRIES = 30
RECONNECT_WAIT = 7


HOST = "127.0.0.1"
PORT = 8080
MODEM_ADDRESS = Address("11.11.11")


def _create_direct_ack(msg):
    """Create a DIRECT ACK response to a standard or extended outbound message."""
    msg_flag = MessageFlags.create(MessageFlagType.DIRECT_ACK, False)
    addr = msg.address
    cmd1 = msg.cmd1
    cmd2 = msg.cmd2
    dir_ack = bytearray([0x02, int(MessageId.STANDARD_RECEIVED)])
    dir_ack.extend(bytes(addr))
    dir_ack.extend(bytes(MODEM_ADDRESS))
    dir_ack.extend(bytes(msg_flag))
    dir_ack.extend(bytearray([cmd1, cmd2]))
    return dir_ack


async def async_connect_mock(protocol, host=None, port=None):
    """Connect to a mock transport."""
    host = HOST if not host else host
    port = PORT if not port else port
    transport = MockTransport(protocol=protocol, host=host, port=port)
    if await transport.async_test_connection():
        _LOGGER.debug("Connection made async_connect_mock")
        transport.start_reader()
        protocol.connection_made(transport)
        return transport
    return None


# This transport is designed for the Hub version 2.
# Hub version 1 should use socket interface on port 9761.
# pylint: disable=too-many-instance-attributes
class MockTransport(asyncio.Transport):
    """An asyncio transport model of a mock communication channel.

    A transport class is an abstraction of a communication channel.
    This allows protocol implementations to be developed against the
    transport abstraction without needing to know the details of the
    underlying channel, such as whether it is a pipe, a socket, or
    indeed an HTTP connection.


    You generally won’t instantiate a transport yourself; instead, you
    will call `async_connect_mock` which will create the
    transport and try to initiate the underlying communication channel,
    calling you back when it succeeds.
    """

    def __init__(self, protocol, host="127.0.0.1", port=8080):
        """Init the HttpTransport class."""
        super().__init__()
        self._protocol = protocol
        self._host = host
        self._port = port
        self._webserver = MockReader(host, port)

        self._closing = False
        self._reader_lock = asyncio.Lock()
        self._read_write_lock = asyncio.Lock()
        self._last_read = asyncio.Queue()
        self._last_msg = None
        self._reader_task = None

    @property
    def write_wait(self):
        """Return the time to wait between writes."""
        return 0.05

    def abort(self):
        """Alternative to closing the transport."""
        self.close()

    def can_write_eof(self):
        """Return False always."""
        return False

    def is_closing(self):
        """Return the state of the transport.

        True if the transport is closed or in the process of closing.

        """
        return self._closing

    def close(self):
        """Close the transport."""
        _LOGGER.debug("Closing Hub session")
        self._closing = True
        self._webserver.stop()

    def get_write_buffer_size(self):
        """Rreturn 0 (i.e. none) always."""
        return 0

    def pause_reading(self):
        """Pause the read."""
        asyncio.ensure_future(self._stop_reader(False))

    def resume_reading(self):
        """Resume the reader."""
        self.start_reader()

    def set_write_buffer_limits(self, high=None, low=None):
        """Not implemented."""
        raise NotImplementedError("HTTP connections do not support write buffer limits")

    def write(self, data):
        """Write data to the transport."""
        asyncio.ensure_future(self.async_write(data))

    async def async_write(self, data):
        """Async write to the transport."""
        await asyncio.sleep(0.01)
        if data.message_id == MessageId.GET_IM_INFO:
            ack_response = bytearray([0x02, int(MessageId.GET_IM_INFO)])
            ack_response.extend(bytearray(bytes(MODEM_ADDRESS)))
            ack_response.append(0x03)
            ack_response.append(0x00)
            ack_response.append(0x00)
            ack_response.append(int(AckNak.ACK))
        elif data.message_id == MessageId.GET_IM_CONFIGURATION:
            ack_response = bytearray(
                [
                    0x02,
                    int(MessageId.GET_IM_CONFIGURATION),
                    0x00,
                    0x00,
                    0x00,
                    AckNak.ACK,
                ]
            )
        else:
            ack_response = bytes(data) + bytes([int(AckNak.ACK)])
        self._protocol.data_received(bytes(ack_response))
        await asyncio.sleep(0.01)
        if data.message_id in [MessageId.SEND_STANDARD, MessageId.SEND_EXTENDED]:
            dir_ack = _create_direct_ack(data)
            self._protocol.data_received(bytes(dir_ack))
            await asyncio.sleep(0.01)

    def start_reader(self):
        """Start the reader."""
        if not self._reader_lock.locked():
            self._closing = False
            self._start_reader()

    async def async_test_connection(self):
        """Test the connection to the hub."""
        return True

    def write_eof(self):
        """Not implemented."""
        raise NotImplementedError("HTTP connections do not support end-of-file")

    def writelines(self, list_of_data):
        """Not implemented."""
        raise NotImplementedError("HTTP connections do not support writelines")

    # pylint: disable=broad-except
    async def _ensure_reader(self):
        _LOGGER.info("Insteon mock reader started")
        await self._webserver.async_start()
        async with self._reader_lock:
            while not self._closing:
                buffer = None
                async with self._read_write_lock:
                    buffer = await self._webserver.queue.get()
                    if buffer:
                        _LOGGER.debug("New buffer: %s", buffer)
                        buffer = self._check_strong_nak(buffer)
                        try:
                            bin_buffer = unhexlify(buffer)
                        except (Error, Incomplete):
                            _LOGGER.warning("Invalid buffer: %s", buffer)
                        else:
                            self._protocol.data_received(bin_buffer)
                if not self._closing:
                    await asyncio.sleep(READ_WAIT)
        _LOGGER.info("Insteon mock reader stopped")

    def _reader_closed_callback(self, exc):
        """Call when the reader closes."""
        self._closing = True
        self._protocol.connection_lost(exc)

    def _check_strong_nak(self, buffer):
        """Check if a NAK message is received with multiple `NAKs`.

        There appears to be a bug in the Hub that produces a series
        of NAK codes rather than returning the original message
        with a single NAK.
        """
        naks = buffer.count("15")
        if (naks >= 9 or naks == len(buffer) / 2) and self._last_msg is not None:
            return self._last_msg.hex() + "15"
        return buffer

    # pylint: disable=unused-argument
    def _start_reader(self, future=None):
        _LOGGER.debug("Starting the buffer reader")
        if not self._closing:
            self._reader_task = asyncio.ensure_future(self._ensure_reader())
            self._reader_task.add_done_callback(self._reader_closed_callback)

    async def _stop_reader(self, reconnect=False):
        _LOGGER.debug("Stopping the reader and reconnect is %s", reconnect)
        self.close()
        if self._reader_lock.locked():
            if not reconnect:
                self._reader_task.remove_done_callback(self._reader_closed_callback)



"""Modem command to get next ALDB record."""
import logging

from . import ack_handler, nak_handler
from ..constants import ResponseStatus
from ..topics import READ_EEPROM
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class ReadEepromHandler(OutboundHandlerBase):
    """Handle Read EEPROM commands."""

    def __init__(self):
        """Init the ReadEepromHandler class."""
        super().__init__(topic=READ_EEPROM)
        _LOGGER.debug("Setup ReadEepromHandler")

    # pylint: disable=arguments-differ
    async def async_send(self, mem_addr: int):
        """Send the Read from EEPROM message."""
        mem_array = mem_addr.to_bytes(2, "big")

        return await super().async_send(mem_hi=mem_array[0], mem_low=mem_array[1] - 7)

    @ack_handler
    async def async_handle_ack(self, mem_hi: int, mem_low: int):
        """Send the Read from EEPROM message."""
        await self._async_handle_ack(mem_hi=mem_hi, mem_low=mem_low)

    @nak_handler
    async def async_handle_nak(self, mem_hi: int, mem_low: int):
        """Receive the NAK message and return False."""
        await self._message_response.put(ResponseStatus.FAILURE)



"""Access control device types."""

from ..events import OFF_EVENT, OFF_FAST_EVENT, ON_EVENT, ON_FAST_EVENT
from ..groups import ON_OFF_SWITCH
from .on_off_responder_base import OnOffResponderBase


class AccessControl_Morningstar(OnOffResponderBase):
    """Access Control MorningStar device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
        state_name=ON_OFF_SWITCH,
        on_event_name=ON_EVENT,
        off_event_name=OFF_EVENT,
        on_fast_event_name=ON_FAST_EVENT,
        off_fast_event_name=OFF_FAST_EVENT,
    ):
        """Init the OnOffResponderBase class."""
        buttons = {1: (ON_OFF_SWITCH, 0)} if buttons is None else buttons
        super().__init__(
            address,
            cat,
            subcat,
            firmware,
            description,
            model,
            buttons,
            on_event_name,
            off_event_name,
            on_fast_event_name,
            off_fast_event_name,
        )

    async def async_lock(self):
        """Lock the device."""
        return await self.async_on()

    async def async_unlock(self):
        """Unlock the device."""
        return await self.async_off()



"""Manage getting an operating flag."""
import asyncio
from collections import namedtuple
import logging
from typing import Dict, Union

from ..address import Address
from ..constants import EngineVersion, ResponseStatus
from ..handlers.to_device.get_operating_flags import GetOperatingFlagsCommand
from ..handlers.to_device.set_operating_flags import SetOperatingFlagsCommand
from ..utils import multiple_status

OperatingFlagInfo = namedtuple("FlagInfo", "name group bit set_cmd unset_cmd")
_LOGGER = logging.getLogger(__name__)
MAX_RETRIES = 5


class GetSetOperatingFlagsManager:
    """Manager to get operating flags."""

    def __init__(self, address: Address, op_flags, engine_version=EngineVersion.I2):
        """Init the GetOperatingFlagsManager class."""
        self._address = Address(address)
        self._op_flags = op_flags
        self._groups: Dict[
            int, Union[OperatingFlagInfo, Dict[int, OperatingFlagInfo]]
        ] = {}
        self._flags: Dict[str, OperatingFlagInfo] = {}
        self._get_command = GetOperatingFlagsCommand(self._address)
        self._set_command = SetOperatingFlagsCommand(self._address)
        self._get_command.subscribe(self._update_flags)
        self._send_lock = asyncio.Lock()
        self._extended_write = False
        self._extended_read = False

    @property
    def extended_write(self):
        """Return if an extended message is required."""
        return self._extended_write

    @extended_write.setter
    def extended_write(self, value: bool):
        """Set the extended write flag."""
        self._extended_write = bool(value)

    def subscribe(self, name, group, bit, set_cmd, unset_cmd):
        """Subscribe to updates."""
        flag_info = OperatingFlagInfo(name, group, bit, set_cmd, unset_cmd)
        if self._groups.get(group) is None:
            self._groups[group] = {}
        if bit is not None:
            self._groups[group][bit] = flag_info
        else:
            self._groups[group] = flag_info
        self._flags[name] = flag_info

    def unsubscribe(self, name):
        """Remove a flag from updates."""
        flag_info = self._flags.get(name)
        if flag_info:
            group = flag_info.group
            bit = flag_info.bit
            self.remove(name, group, bit)

    def check_duplicate(self, name, group, bit, set_cmd, unset_cmd):
        """Check if the Operating Flag is a duplicate."""
        for _, flag in self._flags.items():
            if flag.name == name:
                raise ValueError(f"Adding duplicate operating flag {name}")
            if flag.group == group and flag.bit == bit:
                raise ValueError(
                    f"Adding duplicate operating flag {name} in group {group} and bit {bit}. Existing flag name: {flag.name}"
                )
            if set_cmd is not None and flag.set_cmd == set_cmd:
                raise ValueError(
                    f"Adding duplicate operating flag {name} wit set command {set_cmd}. Existing flag name: {flag.name}"
                )
            if unset_cmd is not None and flag.unset_cmd == unset_cmd:
                raise ValueError(
                    f"Adding duplicate operating flag {name} wit umset command {unset_cmd}. Existing flag name: {flag.name}"
                )

    def remove(self, name, group, bit=None):
        """Remove a flag."""
        try:
            if bit is not None:
                self._groups[group].pop(bit)
            else:
                self._groups.pop(group)
        except KeyError:
            pass

        try:
            self._flags.pop(name)
        except KeyError:
            pass

    async def async_read(self, group=None):
        """Get the operating flags."""
        results = []
        if group is None:
            for curr_group in self._groups:
                result = await self._async_read(curr_group)
                results.append(result)
            return multiple_status(*results)
        return await self._async_read(group)

    async def _async_read(self, group):
        retries = 0
        result = ResponseStatus.UNSENT
        while retries < MAX_RETRIES and result != ResponseStatus.SUCCESS:
            result = await self._get_command.async_send(
                group, extended=self._extended_read
            )
            if result == ResponseStatus.DIRECT_NAK_CHECK_SUM:
                self._extended_read = True
                retries = -1
            elif result in [
                ResponseStatus.DIRECT_NAK_ALDB,
                ResponseStatus.DIRECT_NAK_INVALID_COMMAND,
            ]:
                return result
            await asyncio.sleep(0.01)
            retries += 1
        return result

    async def async_write(self):
        """Set the operating flags."""
        results = []
        for name, flag_info in self._flags.items():
            if self._op_flags[name].is_dirty:
                result = await self._async_write(flag_info)
                results.append(result)
        return multiple_status(*results)

    async def _async_write(self, flag_info):
        flag = self._op_flags[flag_info.name]
        should_set = not flag.new_value if flag.is_reversed else flag.new_value
        cmd = flag_info.set_cmd if should_set else flag_info.unset_cmd
        if cmd is not None:  # The operating flag is read only
            retries = 0
            result = ResponseStatus.UNSENT

            while retries < MAX_RETRIES and result != ResponseStatus.SUCCESS:
                result = await self._set_command.async_send(
                    cmd=cmd, extended=self._extended_write
                )
                if result == ResponseStatus.DIRECT_NAK_CHECK_SUM:
                    self._extended_read = True
                    retries = -1
                elif result in [
                    ResponseStatus.DIRECT_NAK_ALDB,
                    ResponseStatus.DIRECT_NAK_INVALID_COMMAND,
                ]:
                    return result
                await asyncio.sleep(0.01)
                retries += 1

            if result == ResponseStatus.SUCCESS:
                flag.set_value(flag.new_value)
            return result

        # Reset the read only flag to original value
        if flag.is_read_only:
            flag.set_value(flag.value)
        return ResponseStatus.SUCCESS

    def _update_flags(self, group, flags):
        """Update each flag."""

        if not self._groups.get(group):
            return

        if isinstance(self._groups[group], OperatingFlagInfo):
            flag_info = self._groups[group]
            flag = self._op_flags[flag_info.name]
            flag.set_value(flags)
            return

        for bit in self._groups[group]:
            flag_info = self._groups[group][bit]
            flag = self._op_flags[flag_info.name]
            value = bool(flags & 1 << flag_info.bit)
            flag.set_value(value)



"""Send a Start All-Linking command."""
from . import ack_handler, nak_handler
from ..topics import SEND_ALL_LINK_COMMAND
from .outbound_base import OutboundHandlerBase


class SendAllLinkCommandHandler(OutboundHandlerBase):
    """Handle Start All-LInking commands."""

    def __init__(self):
        """Init the StartAllLinking class."""
        super().__init__(SEND_ALL_LINK_COMMAND)

    # pylint: disable=arguments-differ
    async def async_send(self, group: int, cmd1: int, cmd2: int) -> bool:
        """Send the Start All-Linking message asyncronously."""
        return await super().async_send(group=group, cmd1=cmd1, cmd2=cmd2)

    @ack_handler
    async def async_handle_ack(self, group, cmd1, cmd2):
        """Handle the ACK response."""
        await self._async_handle_ack()

    @nak_handler
    async def async_handle_nak(self, group, cmd1, cmd2):
        """Handle the NAK response."""
        await self._async_handle_nak()



"""Base class to handle Broadcast messages from devices."""

from ...constants import MessageFlagType
from ..inbound_base import InboundHandlerBase


class AllLinkCleanupCommandHandlerBase(InboundHandlerBase):
    """Base class to handle inbound All-Link Cleanup messages."""

    def __init__(self, topic, address, group):
        """Init the broadcast_handlerBase class."""

        super().__init__(
            topic=topic,
            address=address,
            group=group,
            message_type=MessageFlagType.ALL_LINK_CLEANUP,
        )



"""Send an X10 command."""

from .. import ack_handler, nak_handler
from ...constants import ResponseStatus, X10Commands, X10CommandType
from ...topics import X10_SEND
from ...x10_address import X10Address
from ..outbound_base import OutboundHandlerBase


class X10CommandSend(OutboundHandlerBase):
    """Send an X10 command."""

    def __init__(self, address, x10_cmd: X10Commands):
        """Send an X10 command."""
        super().__init__(topic=X10_SEND)
        self._address = X10Address(address)
        self._cmd = x10_cmd

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the command."""
        if self._cmd in [
            X10Commands.ALL_LIGHTS_OFF,
            X10Commands.ALL_LIGHTS_ON,
            X10Commands.ALL_UNITS_OFF,
        ]:
            cmd_int = int(self._cmd)
            raw_x10 = (self._address.housecode_byte << 4) + cmd_int
            return await super().async_send(
                raw_x10=raw_x10, x10_flag=X10CommandType.COMMAND
            )

        raw_x10 = bytes(self._address)
        uc_result = await super().async_send(
            raw_x10=raw_x10, x10_flag=X10CommandType.UNITCODE
        )

        if uc_result != ResponseStatus.SUCCESS:
            return uc_result

        cmd_int = int(self._cmd)
        raw_x10 = (self._address.housecode_byte << 4) + cmd_int
        return await super().async_send(
            raw_x10=raw_x10, x10_flag=X10CommandType.COMMAND
        )

    @ack_handler
    async def async_handle_ack(self, raw_x10, x10_flag):
        """Handle the X10 message ACK."""
        await self._async_handle_ack()

    @nak_handler
    async def async_handle_nak(self, raw_x10, x10_flag):
        """Handle the X10 message NAK.

        If the message NAKs it is likely the modem does not support X10 messages.
        The device will try to send the X10 message 3 times to ensure it is not a
        modem ready issue.

        This handler ensures the modem does not try to resend the message constantly.

        """
        await self._async_handle_nak()



"""Cancel All-Linking."""
from . import ack_handler, nak_handler
from ..topics import CANCEL_ALL_LINKING
from .outbound_base import OutboundHandlerBase


# pylint: disable=arguments-differ
class CancelAllLinkingCommandHandler(OutboundHandlerBase):
    """Cancel All-Linking Command."""

    def __init__(self):
        """Init the CancelAllLinkingCommandHandler class."""
        super().__init__(topic=CANCEL_ALL_LINKING)

    @ack_handler
    async def async_handle_ack(self):
        """Handle the ACK response from the modem."""
        return await self._async_handle_ack()

    @nak_handler
    async def async_handle_nak(self):
        """Handle the ACK response from the modem."""
        return await self._async_handle_nak()



"""Manage outbound ON command to a device."""

from ...topics import ON
from .broadcast_command import BroadcastCommandHandlerBase


class OnLevelInbound(BroadcastCommandHandlerBase):
    """On Level command inbound."""

    def __init__(self, address, group):
        """Init the OnLevelInbound class."""
        super().__init__(topic=ON, address=address, group=group)

    def _handle_message_received(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the ON command from a device."""
        self._call_subscribers(on_level=cmd2 if cmd2 else 0xFF)



"""Extended property set manager."""

from typing import Dict

from ..address import Address
from ..config.extended_property import ExtendedProperty
from ..constants import ResponseStatus
from ..handlers.to_device.extended_set import ExtendedSetCommand

RETRIES = 3


def _value(prop: ExtendedProperty):
    """Return the usable value from a property."""
    return prop.new_value if prop.is_dirty else prop.value


class ExtendedPropertyWriteManager:
    """Extended property read manager."""

    def __init__(
        self,
        address: Address,
        properties: Dict[int, ExtendedProperty],
        data2: int,
        cmd2: int = 0,
        data1: int = 0x00,
    ):
        """Init the ExtendedPropertyReadManager class."""
        self._address = Address(address)
        self._write_command = ExtendedSetCommand(
            self._address, data1=data1, data2=data2, cmd2=cmd2
        )
        self._props: Dict[int, ExtendedProperty] = properties

    @property
    def is_dirty(self):
        """Return if any properties are dirty."""
        for _, prop in self._props.items():
            if prop.is_dirty:
                return True
        return False

    async def async_write(self):
        """Write the extended properties."""
        if not self.is_dirty:
            return ResponseStatus.SUCCESS

        ud_dict = {f"data{index}": _value(prop) for index, prop in self._props.items()}
        retries = RETRIES
        response = ResponseStatus.UNSENT
        while retries and response != ResponseStatus.SUCCESS:
            response = await self._write_command.async_send(**ud_dict)
            retries -= 1
            if response == ResponseStatus.SUCCESS:
                self._load_prop_values()
        return response

    def _load_prop_values(self):
        """Load the new value to the property value."""
        for _, prop in self._props.items():
            if prop.is_dirty:
                prop.set_value(prop.new_value)



"""Modem command to get next ALDB record."""
import logging

from . import ack_handler, nak_handler
from ..constants import ResponseStatus
from ..topics import GET_NEXT_ALL_LINK_RECORD
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class GetNextAllLinkRecordHandler(OutboundHandlerBase):
    """Handle Get Next All Link Record commands."""

    def __init__(self):
        """Init the GetNextAldbRecordNak class."""
        super().__init__(topic=GET_NEXT_ALL_LINK_RECORD)

    @ack_handler
    async def async_handle_ack(self):
        """Handle the ACK response from the modem."""
        return await self._async_handle_ack()

    @nak_handler
    async def async_handle_nak(self):
        """Handle the NAK response from the modem."""
        await self._message_response.put(ResponseStatus.FAILURE)



"""Manage wet / dry state of the leak sensor."""
from typing import Callable

from ..address import Address
from ..handlers.from_device.on_level import OnLevelInbound
from ..subscriber_base import SubscriberBase


class WetDryManager(SubscriberBase):
    """Manage the wet / dry state of a leak sensor."""

    class WetDryEvent(SubscriberBase):
        """Wet or dry events."""

        def call_subscribers(self, dry):
            """Call subscribers of wet/dry events."""
            self._call_subscribers(dry=dry)

    def __init__(self, address, dry_group=1, wet_group=2):
        """Init the WetDryManager."""
        self._address = Address(address)
        self._wet_group = wet_group
        self._dry_group = dry_group
        subscriber_topic = f"subscriber_{self._address.id}_wet_dry"
        super().__init__(subscriber_topic)

        self._dry_handler = OnLevelInbound(self._address, self._dry_group)
        self._dry_events = self.WetDryEvent(f"{subscriber_topic}_dry")
        self._dry_handler.subscribe(self._dry)

        self._wet_handler = OnLevelInbound(self._address, self._wet_group)
        self._wet_events = self.WetDryEvent(f"{subscriber_topic}_wet")
        self._wet_handler.subscribe(self._wet)

    def subscribe(self, callback: Callable, force_strong_ref=False):
        """Subscribe to the event."""
        if force_strong_ref and callback not in self._subscribers:
            self._subscribers.append(callback)
        self.subscribe_wet(callback)
        self.subscribe_dry(callback)

    def subscribe_dry(self, callback):
        """Subscribe to dry events."""
        self._dry_events.subscribe(callback)

    def subscribe_wet(self, callback):
        """Subscribe to wet events."""
        self._wet_events.subscribe(callback)

    def _dry(self, on_level):
        """Dry event received."""
        self._dry_events.call_subscribers(dry=True)

    def _wet(self, on_level):
        """Dry event received."""
        self._wet_events.call_subscribers(dry=False)



"""Dimmable Lighting Control Devices (CATEGORY 0x01)."""

from ..handlers.to_device.off import OffCommand
from ..handlers.to_device.off_fast import OffFastCommand
from ..handlers.to_device.on_fast import OnFastCommand
from ..handlers.to_device.on_level import OnLevelCommand
from .device_commands import (
    OFF_COMMAND,
    OFF_FAST_COMMAND,
    ON_COMMAND,
    ON_FAST_COMMAND,
    STATUS_COMMAND,
)
from .variable_controller_base import VariableControllerBase


class VariableResponderBase(VariableControllerBase):
    """Variable Responder Base Device."""

    def on(self, on_level=0xFF, group=0, fast=False):
        """Turn on the device.

        Parameters:
            on_level: Default 0xff (full on). Set the device on level.
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        """
        group = 1 if not group else group
        command = ON_FAST_COMMAND if fast else ON_COMMAND
        self._handlers[group][command].send(on_level=on_level)

    async def async_on(self, on_level=0xFF, group=0, fast=False):
        """Turn on the device.

        Parameters:
            on_level: Default 0xff (full on). Set the device on level.
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        group = 1 if not group else group
        command = ON_FAST_COMMAND if fast else ON_COMMAND
        return await self._handlers[group][command].async_send(on_level)

    def off(self, group=0, fast=False):
        """Turn off the device.

        Parameters:
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        """
        group = 1 if not group else group
        command = OFF_FAST_COMMAND if fast else OFF_COMMAND
        self._handlers[group][command].send()

    async def async_off(self, group=0, fast=False):
        """Turn off the device.

        Parameters:
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        group = 1 if not group else group
        command = OFF_FAST_COMMAND if fast else OFF_COMMAND
        return await self._handlers[group][command].async_send()

    async def async_status(self, group=None):
        """Get the device status."""
        status_type = self._groups[group].status_type if group is not None else None
        return await self._managers[STATUS_COMMAND].async_status(status_type)

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        for group in self._buttons:
            if self._handlers.get(group) is None:
                self._handlers[group] = {}
            self._handlers[group][ON_COMMAND] = OnLevelCommand(self._address, group)
            self._handlers[group][OFF_COMMAND] = OffCommand(self._address, group)
            self._handlers[group][ON_FAST_COMMAND] = OnFastCommand(self._address, group)
            self._handlers[group][OFF_FAST_COMMAND] = OffFastCommand(
                self._address, group
            )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        for group in self._buttons:
            state = self._groups[group]

            self._handlers[group][ON_COMMAND].subscribe(state.set_value)
            self._handlers[group][OFF_COMMAND].subscribe(state.set_value)
            self._handlers[group][ON_FAST_COMMAND].subscribe(state.set_value)
            self._handlers[group][OFF_FAST_COMMAND].subscribe(state.set_value)



"""Manage the ALDB write process for an Insteon Modem."""

import logging
from typing import Tuple

from ..aldb.aldb_record import ALDBRecord
from ..constants import ManageAllLinkRecordAction, ReadWriteMode, ResponseStatus
from ..handlers.manage_all_link_record import ManageAllLinkRecordCommand
from ..handlers.write_eeprom import WriteEepromHandler
from .aldb_write_manager import ALDBWriteException

_LOGGER = logging.getLogger(__name__)
MAX_RETRIES = 5


class ImWriteManager:
    """Manage writes to the Insteon Modem All-Link Database."""

    def __init__(self, aldb):
        """Init the ImWriteManager."""
        self._write_cmd = ManageAllLinkRecordCommand()
        self._aldb = aldb

    @property
    def can_write(self):
        """Return true if the AlDB is loaded."""
        return self._aldb.is_loaded

    async def async_write(
        self, record, force=False
    ) -> Tuple[ResponseStatus, ALDBRecord]:
        """Write a record to the ALDB."""

        if not self.can_write and not force:
            raise ALDBWriteException(
                "All-Link databased must be loaded before writing."
            )
        if self._aldb.read_write_mode == ReadWriteMode.EEPROM:
            return await self.async_write_record(record)

        if not record.is_in_use:
            return await self._async_write_delete_standard(record)

        return await self._async_write_change_standard(record)

    async def async_write_record(self, record: ALDBRecord):
        """Write to EEPROM."""
        cmd = WriteEepromHandler()
        retries = 5
        result = ResponseStatus.UNSENT
        while retries and result != ResponseStatus.SUCCESS:
            result = await cmd.async_send(
                mem_addr=record.mem_addr,
                controller=record.is_controller,
                group=record.group,
                target=record.target,
                data1=record.data1,
                data2=record.data2,
                data3=record.data3,
                in_use=record.is_in_use,
                high_water_mark=record.is_high_water_mark,
                bit5=record.is_bit5_set,
                bit4=record.is_bit4_set,
            )
            retries -= 1
        return result

    async def _async_write_change_standard(
        self, record: ALDBRecord
    ) -> Tuple[ResponseStatus, ALDBRecord]:
        """Write a changed record."""
        curr_record = self._aldb.get(record.mem_addr, record)

        # Cannot change a controller to a responder or vise versa so delete first
        if curr_record.is_controller != record.is_controller:
            await self._async_write_delete_standard(curr_record)
            curr_record = record

        if curr_record.is_controller:
            action = ManageAllLinkRecordAction.MOD_FIRST_CTRL_OR_ADD
        else:
            action = ManageAllLinkRecordAction.MOD_FIRST_RESP_OR_ADD

        result = await self._async_write_record_standard(action, record)

        # If the record is already identical to `record` the modem returns a NAK
        # So test if a matching record is in the modem's ALDB
        if result != ResponseStatus.SUCCESS:
            if await self._find_matching_record(record):
                return ResponseStatus.SUCCESS
        return result

    async def _async_write_delete_standard(
        self, record
    ) -> Tuple[ResponseStatus, ALDBRecord]:
        """Write a deleted record."""
        matching_recs = await self._aldb.async_find_records(
            target=record.target, group=record.group
        )
        rec_to_delete = None
        for rec in matching_recs:
            if rec.is_controller == record.is_controller:
                rec_to_delete = rec
                break
        if rec_to_delete:
            return self._async_write_record_standard(
                rec_to_delete, ManageAllLinkRecordAction.DELETE_FIRST
            )

    async def _async_write_record_standard(
        self, action, record
    ) -> Tuple[ResponseStatus, ALDBRecord]:
        """Perform the ALDB write action."""
        retries = 0
        response = ResponseStatus.UNSENT
        while response != ResponseStatus.SUCCESS and retries < MAX_RETRIES:
            response = await self._write_cmd.async_send(
                action=action,
                controller=record.is_controller,
                group=record.group,
                target=record.target,
                data1=record.data1,
                data2=record.data2,
                data3=record.data3,
                in_use=record.is_in_use,
                high_water_mark=record.is_high_water_mark,
                bit5=record.is_bit5_set,
                bit4=record.is_bit4_set,
            )
            retries += 1

        return response

    async def _find_matching_record(self, record: ALDBRecord):
        """Find a matching record in the ALDB."""
        async for rec in self._aldb.async_find_records(record.target, record.group):
            if record.is_exact_match(rec):
                return rec
        return None



"""Device configuration settings."""

from ..constants import ToggleMode

# Used to hide a bit to ensure we set it correctly in a write command
# when used in an Extended Property. For example, if an extended
# property write requires a data field but that data field is undefined
# we can create a UNUSED extended properity that will be used only
# to rewrite the existing value if any.
UNUSED = "unused"

# Operating Flags
PROGRAM_LOCK_ON = "program_lock_on"
BUTTON_LOCK_ON = "button_lock_on"
LED_ON = "led_on"
LED_OFF = "led_off"
KEY_BEEP_ON = "key_beep_on"
STAY_AWAKE_ON = "stay_awake_on"
LISTEN_ONLY_ON = "listen_only_on"
HEART_BEAT_ON = "heart_beat_on"
HEART_BEAT_OFF = "heart_beat_off"
SEND_ON_ONLY = "send_on_only_on"
GROUPED_ON = "grouped_on"
TWO_GROUPS_ON = "two_groups_on"
RESUME_DIM_ON = "resume_dim_on"
RF_DISABLE_ON = "rf_disable_on"
RELAY_AT_FULL_ON = "relay_at_full_on"
POWERLINE_DISABLE_ON = "powerline_disable_on"
TEND_ON = "tend_on"
X10_OFF = "x10_off"
CLEANUP_REPORT_ON = "cleanup_report_on"
CLEANUP_REPORT_OFF = "cleanup_report_off"
DETACH_LOAD_ON = "detach_load_on"
SMART_HOPS_ON = "smart_hops_on"
LED_BLINK_ON_ERROR_ON = "blink_on_error_on"
LED_BLINK_ON_ERROR_OFF = "blink_on_error_off"
LED_BLINK_ON_TX_ON = "blink_on_tx_on"
REPEAT_OPEN_ON = "repeat_open_on"
REPEAT_CLOSED_ON = "repeat_closed_on"
LINK_TO_FF_GROUP = "link_to_ff_group"
LOAD_SENSE_ON = "load_sense_on"
LOAD_SENSE_2_ON = "load_sense_2_on"
IGNORE_JUMPER_ON = "ignore_jumper_on"
NIGHT_MODE_ONLY = "night_mode_only"
MULTI_SEND_ON = "multi_send_on"
SOFTWARE_SUPPORT_ON = "software_support_on"
HARDWARE_SEND_ON_ONLY = "hardware_send_on_only"
HARDWARE_NIGHT_MODE = "hardware_night_mode"
HARDWARE_LED_OFF = "hardware_led_off"
RELAY_ON_SENSE_ON = "relay_on_sense_on"
MOMENTARY_MODE_ON = "momentary_on"
MOMENTARY_ON_OFF_TRIGGER = "momentary_on_off_trigger"
SENSE_SENDS_OFF = "sense_sends_off"
MOMENTARY_FOLLOW_SENSE = "momentary_follow_sense"
INSTEON_OFF = "insteon_off"
CHECKSUM_OFF = "checksum_off"
STANDARD_HOLDOFF = "standard_holdoff"
DUAL_LINE_ON = "dual_line_on"
LATCHING_ON = "latching_on"
THREE_WAY_ON = "three_way_on"
REVERSED_ON = "reversed_on"
FORWARD_ON = "forward_on"
MOMENTARY_LINE_ON = "momentary_line_on"
NOT_3_WAY = "not_3_way"
CRC_ERROR_COUNT = "crc_error_count"
SIGNAL_TO_NOISE_OF_LAST_FAILURE = "signal_to_noise_of_last_failure"
SIGNAL_TO_NOISE_FAILURE_COUNT = "signal_to_noise_failure_count"
DATABASE_DELTA = "database_delta"
CELSIUS = "celsius"
TIME_24_HOUR_FORMAT = "time_24_hour_format"
SOFTWARE_LOCK_ON = "software_lock_on"
NIGHT_MODE_ON = "night_mode_on"
DISABLE_BUTTONS = "disable_buttons"
ACK_A_SCENE = "ack_a_scene"
YAKETY_YAK = "yakety_yak"
RED_LED_OFF = "red_led_off"
GREEN_LED_OFF = "green_led_off"
BLUE_LED_OFF = "blue_led_off"
SKIP_SOME_HOPS = "skip_some_hops"
NO_CACHE = "no_cache"
DO_NOT_ROTATE_TO_OFF = "do_not_rotate_to_off"
USE_RELAY_AT_FF = "use_relay_at_ff"
USE_LOCAL_PROFILE = "use_local_profile"
MOD_SYNC_SYNC = "mod_sync_sync"
NIGHT_LEDS_ON = "night_leds_on"
RELAY_MODE_OFF = "relay_mode_off"

# Extended Properties
RAMP_RATE = "ramp_rate"
X10_HOUSE = "x10_house"
X10_UNIT = "x10_unit"
LED_DIMMING = "led_dimming"
ON_LEVEL = "on_level"
AWAKE_INTERVAL = "awake_interval"
SLEEP_INTERVAL = "sleep_interval"
BROADCAST_NUMBER = "broadcast_number"
TRIGGER_GROUP_MASK = "trigger_group_mask"
LSB_OF_SLEEP_INTERVAL = "lsb_of_sleep_interval"
APP_RETRIES = "app_retries"
CONFIG = "config"
BATTERY_LEVEL = "battery_level"
DATABASE_DELTA = "database_delta"
SENSOR_STATUS = "sensor_status"
HEARBEAT_INTERVAL = "heartbeat_interval"
BATTERY_LOW_LEVEL = "battery_low_level"
LED_BRIGHTNESS = "led_brightness"
MOTION_TIMEOUT = "motion_timeout"
LIGHT_SENSITIVITY = "light_sensitivity"
HARDWARE_TIMEOUT = "hardware_timeout"
HARDWARE_LIGHT_SENSITIVITY = "hardware_light_sensitivity"
AMBIENT_LIGHT_INTENSITY = "ambient_light_intensity"
DELAY = "delay"
PRESCALER = "prescaler"
DURATION_HIGH = "duration_high"
DURATION_LOW = "duration_low"
ON_MASK = "on_mask"
OFF_MASK = "off_mask"
NON_TOGGLE_MASK = "non_toggle_mask"
NON_TOGGLE_ON_OFF_MASK = "non_toggle_on_off_mask"
GROUP_MASK = "group_mask"
BACKLIGHT = "backlight"
CHANGE_DELAY = "change_delay"
MASTER = "master"
HUMIDITY_HIGH = "humidity_high"
HUMIDITY_LOW = "humidity_low"
TEMP_OFFSET = "temp_offset"
TEMP_OFFSET_EXTERNAL = "temp_offset_external"
HUMIDITY_OFFSET = "humidity_offset"
OPERATING_FLAGS = "operating_flags"
MAX_RETRIES = "max_retries"
BRIGHTNESS_MAX = "brightness_max"
BRIGHTNESS_MIN = "brightness_min"
BRIGHTNESS_START = "brightness_start"
TEMP_HOT = "temp_hot"
TEMP_VERY_HOT = "temp_very_hot"
NIGHT_MODE_MAX_LEVEL = "night_mode_max_level"
NIGHT_MODE_LED_BRIGHTNESS = "night_mode_led_brightness"
NIGHT_MODE_RAMP_RATE = "night_mode_ramp_rate"
LOAD_BUTTON_NUMBER = "load_button_number"

# Derived Properties
MOMENTARY_DELAY = "momentary_delay"
RADIO_BUTTON_GROUPS = "radio_button_groups"
RAMP_RATE_IN_SEC = "ramp_rate_in_seconds"
RELAY_MODE = "relay_mode"
TOGGLE_BUTTON = "toggle_button"
LOAD_BUTTON = "load_button"

# Modem properties
MODEM_CONFIG = "modem_config"
DISABLE_AUTO_LINKING = "disable_auto_linking"
MONITOR_MODE = "monitor_mode"
AUTO_LED = "auto_led"
DEADMAN = "deadman"


def calc_toggle_mode(toggle_value, toggle_on_value):
    """Calculate the toggle mode value."""
    if not toggle_value:
        return ToggleMode.TOGGLE
    if toggle_on_value:
        return ToggleMode.ON_ONLY
    return ToggleMode.OFF_ONLY


def get_usable_value(prop):
    """Return the current or new value of the property."""
    return prop.new_value if prop.is_dirty else prop.value



"""Insteon All-Link Database.

The All-Link database contains database records that represent links to other
Insteon devices that either respond to or control the current device.
"""
import asyncio
import logging

from ..constants import ALDBStatus, EngineVersion, ReadWriteMode
from ..managers.aldb_read_manager import ALDBReadManager
from .aldb_base import ALDBBase

_LOGGER = logging.getLogger(__name__)


class ALDB(ALDBBase):
    """All-Link Database for a device."""

    def __init__(
        self,
        address,
        version=EngineVersion.UNKNOWN,
        mem_addr=0x0FFF,
    ):
        """Init the ALDB class."""
        super().__init__(address=address, version=version, mem_addr=mem_addr)
        self._read_manager = ALDBReadManager(self._address, self._mem_addr)

    # pylint: disable=arguments-differ
    async def async_load(
        self, mem_addr: int = 0x00, num_recs: int = 0x00, refresh: bool = False
    ):
        """Load the All-Link Database."""
        _LOGGER.debug("Loading the ALDB async")
        self._update_status(ALDBStatus.LOADING)
        if refresh:
            self.clear()
        else:
            # Pop any unused records to make sure we query them
            unused = list(self.find(in_use=False))
            for rec in unused:
                self._records.pop(rec.mem_addr)

        if self._read_write_mode == ReadWriteMode.UNKNOWN:
            mode = ReadWriteMode.STANDARD
        else:
            mode = self._read_write_mode

        if self._mem_addr not in self._records:
            # Query the first record
            try:
                async for rec in self._read_manager.async_read(
                    mem_addr=0, num_recs=1, read_write_mode=mode
                ):
                    self._mem_addr = rec.mem_addr
                    self._add_record(rec)
            finally:
                await self._read_manager.async_stop()

        try:
            async for rec in self._read_manager.async_read(
                mem_addr=mem_addr,
                num_recs=num_recs,
                read_write_mode=mode,
            ):
                self._add_record(rec)
                await asyncio.sleep(0.1)
                if self._read_write_mode == ReadWriteMode.UNKNOWN:
                    self._read_write_mode = mode
                if self._is_loaded():
                    break
        finally:
            await self._read_manager.async_stop()

        if not self._is_loaded() and num_recs != 0:
            # Loading all records did not work so now we read individual missing records
            next_record = self._calc_next_record()
            while next_record:
                async for rec in self._read_manager.async_read(
                    mem_addr=next_record, num_recs=1
                ):
                    self._add_record(rec)
                prev_record = next_record
                next_record = self._calc_next_record()
                if next_record == prev_record:
                    # The ALDB did not return the requested record so stop
                    break

        if (
            not self._records
            and self._read_write_mode == ReadWriteMode.STANDARD
            and self._version not in [EngineVersion.I2CS, EngineVersion.OTHER]
        ):
            self._read_write_mode = ReadWriteMode.PEEK_POKE
            return await self.async_load(
                mem_addr=mem_addr, num_recs=num_recs, refresh=refresh
            )

        self.set_load_status()

        return self._status

    def _add_record(self, record) -> bool:
        """Add a record to the record set."""
        _LOGGER.debug("Loading record: %s", str(record))
        # Make sure the records make sense
        if (
            self.high_water_mark_mem_addr
            and record.mem_addr < self.high_water_mark_mem_addr
        ):
            _LOGGER.debug("Record is after the HWM: %s", str(record))
            return False

        # If an existing record will be replaced notify of change
        old_record = self._records.get(record.mem_addr)

        # If the old rec is identical to the new rec, do nothing
        if old_record and record.is_exact_match(old_record):
            _LOGGER.debug("Record has not changed:")
            _LOGGER.debug("Old: %s", str(old_record))
            _LOGGER.debug("New: %s", str(record))
            return False

        if old_record and old_record.is_in_use:
            self._notify_change(self._records[record.mem_addr], force_delete=True)

        self._records[record.mem_addr] = record
        self._notify_change(record)
        return True

    def _calc_next_record(self) -> int:
        """Calculate the memory address of the next missing record."""
        if not self._records:
            return self._mem_addr
        last_addr = list(self)[-1]
        if last_addr == self._mem_addr:
            return last_addr - 8

        for mem_addr in range(self._mem_addr, last_addr, -8):
            try:
                rec = self._records[mem_addr]
                if mem_addr == last_addr and rec.is_high_water_mark:
                    return None
            except IndexError:
                return mem_addr

        return last_addr - 8



"""Handle sending a read request for ALDB records."""
import logging

from .. import ack_handler
from ...address import Address
from ...topics import EXTENDED_GET_SET
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class ExtendedGetCommand(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address, cmd2=0x00, data1=None, data2=0x00, data3=None):
        """Init the ExtendedGetCommand."""
        super().__init__(topic=EXTENDED_GET_SET, address=address)
        self._cmd2 = cmd2
        self._data1 = data1
        self._data2 = data2
        self._data3 = data3

    # pylint: disable=arguments-differ
    async def async_send(self, group=0, crc=False):
        """Send Get Operating Flags message asyncronously."""
        self._data1 = group
        response = await super().async_send(
            cmd2=self._cmd2,
            data1=self._data1,
            data2=self._data2,
            data3=self._data3,
            crc=crc,
        )
        self._data1 = None
        return response

    @ack_handler
    async def async_handle_ack(self, cmd1, cmd2, user_data):
        """Handle the ACK response.

        Required to ensure only GET requests are triggered.
        """
        # pylint: disable=too-many-boolean-expressions
        if (
            cmd2 == self._cmd2
            and (self._data1 is None or user_data["d1"] == self._data1)
            and (self._data2 is None or user_data["d2"] == self._data2)
            and (self._data3 is None or user_data["d3"] == self._data3)
        ):
            await super().async_handle_ack(cmd1=cmd1, cmd2=cmd2, user_data=user_data)



"""Manage writes to the All-Link Database."""
import asyncio
from typing import Tuple

from ..aldb.aldb_record import ALDBRecord
from ..constants import ReadWriteMode, ResponseStatus
from ..handlers.to_device.write_aldb import WriteALDBCommandHandler
from ..managers.peek_poke_manager import get_peek_poke_manager


class ALDBWriteException(Exception):
    """All-Link Database write excetion."""


class ALDBWriteManager:
    """Manage writes to the All-Link Database."""

    def __init__(self, aldb):
        """Init the ALDBWriteManager class."""
        self._aldb = aldb
        self._address = aldb.address
        self._write_handler = WriteALDBCommandHandler(self._address)
        self._poke_manager = get_peek_poke_manager(self._aldb.address)

    @property
    def can_write(self):
        """Return if the ALDB is loaded and we can write to it."""
        return self._aldb.is_loaded

    def write(self, record: ALDBRecord):
        """Write the ALDB record."""
        asyncio.ensure_future(self.async_write(record))

    async def async_write(
        self, record: ALDBRecord, force: bool = False
    ) -> Tuple[ResponseStatus, ALDBRecord]:
        """Write the ALDB record."""

        if not self.can_write and not force:
            raise ALDBWriteException(
                "All-Link databased must be loaded before writing."
            )

        if self._aldb.read_write_mode == ReadWriteMode.PEEK_POKE:
            return await self._poke_record(record)
        return await self._write_standard(record=record)

    async def _write_standard(self, record: ALDBRecord):
        """Write a record using standard method."""
        result = ResponseStatus.UNSENT
        retries = 5
        while result != ResponseStatus.SUCCESS and retries:
            result = await self._write_handler.async_send(
                mem_addr=record.mem_addr,
                controller=record.is_controller,
                group=record.group,
                target=record.target,
                data1=record.data1,
                data2=record.data2,
                data3=record.data3,
                in_use=record.is_in_use,
                bit5=record.is_bit5_set,
                bit4=record.is_bit4_set,
            )
            retries -= 1
        return result

    async def _poke_record(self, record: ALDBRecord):
        """Write the record using poke commands."""

        rec_bytes = bytearray(
            [
                record.data3,
                record.data2,
                record.data1,
                record.target.low,
                record.target.middle,
                record.target.high,
                record.group,
                record.control_flags,
            ]
        )
        mem_addr = record.mem_addr
        for curr_byte in range(0, 8):
            result = await self._async_write_byte(
                mem_addr - curr_byte, rec_bytes[curr_byte]
            )
            if result != ResponseStatus.SUCCESS:
                return result
        return ResponseStatus.SUCCESS

    async def _async_write_byte(self, mem_addr, value):
        """Write the byte to memory."""
        retries = 5
        while retries:
            result = await self._poke_manager.async_poke(mem_addr=mem_addr, value=value)
            if result == ResponseStatus.SUCCESS:
                return ResponseStatus.SUCCESS
            retries -= 1
        return ResponseStatus.FAILURE
